"""Tests for tb01nd - Observer Hessenberg form reduction."""

import numpy as np
import pytest
from numpy.testing import assert_allclose


def test_tb01nd_upper_hessenberg_html_example():
    """
    Test upper observer Hessenberg form using HTML doc example.

    N=5, P=3, JOBU='N', UPLO='U'
    Input A: 5x5, C: 3x5
    Validates mathematical properties since Householder signs can vary.
    """
    from slicot import tb01nd

    # Input A: 5x5 from HTML doc - read column-wise: ((A(I,J), I=1,N), J=1,N)
    a = np.array([
        [15.0, 20.0,  4.0,  5.0,  5.0],
        [21.0,  1.0,  1.0,  6.0, 11.0],
        [-3.0,  2.0,  7.0, 12.0, 17.0],
        [ 3.0,  8.0, 13.0, 13.0, -7.0],
        [ 9.0,  9.0, 14.0, -6.0, -1.0]
    ], order='F', dtype=float)

    # Input C: 3x5 from HTML doc - read row-wise: ((C(I,J), J=1,N), I=1,P)
    c = np.array([
        [ 7.0, -1.0,  3.0, -6.0, -3.0],
        [ 4.0,  5.0,  6.0, -2.0, -3.0],
        [ 9.0,  8.0,  5.0,  2.0,  1.0]
    ], order='F', dtype=float)

    a_orig = a.copy()
    c_orig = c.copy()
    n, p = 5, 3

    # Expected output from HTML doc (printed row-by-row)
    # Transformed A (upper observer Hessenberg form)
    a_expected = np.array([
        [  7.1637, -0.9691, -16.5046,  0.2869,  0.9205],
        [ -2.3285, 11.5431,  -8.7471,  3.4122, -3.7118],
        [-10.5440, -7.6032,  -0.3215,  3.6571, -0.4335],
        [ -3.6845,  5.6449,   0.5906,-15.6996, 17.4267],
        [  0.0000, -6.4260,   1.5591, 14.4317, 32.3143]
    ], order='F', dtype=float)

    # Transformed C
    c_expected = np.array([
        [ 0.0000,  0.0000,  7.6585,  5.2973, -4.1576],
        [ 0.0000,  0.0000,  0.0000,  5.8305, -7.4837],
        [ 0.0000,  0.0000,  0.0000,  0.0000,-13.2288]
    ], order='F', dtype=float)

    # Call tb01nd with JOBU='N' (don't compute U)
    a_out, c_out, u_out, info = tb01nd('N', 'U', a, c)

    assert info == 0

    # Verify against expected values from HTML doc (4 decimal places)
    assert_allclose(a_out, a_expected, rtol=1e-3, atol=1e-4)
    assert_allclose(c_out, c_expected, rtol=1e-3, atol=1e-4)


def test_tb01nd_upper_with_u_matrix():
    """
    Test upper observer Hessenberg form with U matrix (JOBU='I').

    Validates mathematical properties: U'AU similarity and CU transformation.
    Random seed: 42 (for reproducibility)
    """
    from slicot import tb01nd

    np.random.seed(42)
    n, p = 5, 3

    a = np.random.randn(n, n).astype(float, order='F')
    c = np.random.randn(p, n).astype(float, order='F')
    a_orig = a.copy()
    c_orig = c.copy()

    # Call with JOBU='I' to get transformation matrix
    a_out, c_out, u_out, info = tb01nd('I', 'U', a, c)

    assert info == 0

    # Verify U is orthogonal: U^T @ U = I
    assert_allclose(u_out.T @ u_out, np.eye(n), rtol=1e-14, atol=1e-14)

    # Verify similarity transformation: A_out = U' @ A_orig @ U
    assert_allclose(a_out, u_out.T @ a_orig @ u_out, rtol=1e-14, atol=1e-14)

    # Verify: C_out = C_orig @ U
    assert_allclose(c_out, c_orig @ u_out, rtol=1e-14, atol=1e-14)

    # Verify upper triangular structure of C (zeros in lower-left)
    # For UPLO='U': C has upper triangular form in last P columns
    # C(i,j)=0 for j < N-P+i for i=0..P-1
    # i=0: j<N-P=2 => C[0,0:2]=0
    # i=1: j<N-P+1=3 => C[1,0:3]=0
    # i=2: j<N-P+2=4 => C[2,0:4]=0
    for i in range(p):
        for j in range(n - p + i):
            assert_allclose(c_out[i, j], 0.0, atol=1e-14)

    # Verify eigenvalues preserved
    eig_before = np.linalg.eigvals(a_orig)
    eig_after = np.linalg.eigvals(a_out)
    assert_allclose(sorted(eig_before.real), sorted(eig_after.real), rtol=1e-13, atol=1e-14)


def test_tb01nd_lower_hessenberg():
    """
    Test lower observer Hessenberg form (UPLO='L').

    Random seed: 123 (for reproducibility)
    """
    from slicot import tb01nd

    np.random.seed(123)
    n, p = 5, 3

    a = np.random.randn(n, n).astype(float, order='F')
    c = np.random.randn(p, n).astype(float, order='F')
    a_orig = a.copy()
    c_orig = c.copy()

    # Call with JOBU='I', UPLO='L'
    a_out, c_out, u_out, info = tb01nd('I', 'L', a, c)

    assert info == 0

    # Verify U is orthogonal: U^T @ U = I
    assert_allclose(u_out.T @ u_out, np.eye(n), rtol=1e-14, atol=1e-14)

    # Verify similarity transformation: A_out = U' @ A_orig @ U
    assert_allclose(a_out, u_out.T @ a_orig @ u_out, rtol=1e-14, atol=1e-14)

    # Verify: C_out = C_orig @ U
    assert_allclose(c_out, c_orig @ u_out, rtol=1e-14, atol=1e-14)

    # For lower observer Hessenberg with UPLO='L':
    # C has lower triangular structure in first P columns
    # C(i,j) = 0 for j > i
    for i in range(p):
        for j in range(i + 1, p):
            assert_allclose(c_out[i, j], 0.0, atol=1e-14)


def test_tb01nd_accumulate_u():
    """
    Test U matrix accumulation mode (JOBU='U').

    When JOBU='U', the given U is updated with the transformations.
    U_out = U_init @ Q where Q is the internal Hessenberg reduction transformation.

    Random seed: 456 (for reproducibility)
    """
    from slicot import tb01nd

    np.random.seed(456)
    n, p = 5, 2

    a = np.random.randn(n, n).astype(float, order='F')
    c = np.random.randn(p, n).astype(float, order='F')
    a_orig = a.copy()
    c_orig = c.copy()

    # Create initial orthogonal U from QR decomposition
    q, _ = np.linalg.qr(np.random.randn(n, n))
    u_init = np.asfortranarray(q.astype(float))
    u_init_copy = u_init.copy()

    # Call with JOBU='U' to update the given U
    a_out, c_out, u_out, info = tb01nd('U', 'U', a, c, u=u_init)

    assert info == 0

    # Verify U_out is still orthogonal
    assert_allclose(u_out.T @ u_out, np.eye(n), rtol=1e-14, atol=1e-14)

    # Extract internal transformation Q = U_init^T @ U_out
    q_transform = u_init_copy.T @ u_out

    # Verify Q is orthogonal
    assert_allclose(q_transform.T @ q_transform, np.eye(n), rtol=1e-14, atol=1e-14)

    # Verify: A_out = Q^T @ A_orig @ Q
    assert_allclose(a_out, q_transform.T @ a_orig @ q_transform, rtol=1e-14, atol=1e-14)

    # Verify: C_out = C_orig @ Q
    assert_allclose(c_out, c_orig @ q_transform, rtol=1e-14, atol=1e-14)


def test_tb01nd_eigenvalue_preservation():
    """
    Validate eigenvalue preservation under similarity transformation.

    Random seed: 789 (for reproducibility)
    """
    from slicot import tb01nd

    np.random.seed(789)
    n, p = 6, 3

    a = np.random.randn(n, n).astype(float, order='F')
    c = np.random.randn(p, n).astype(float, order='F')

    # Compute eigenvalues before transformation
    eig_before = np.linalg.eigvals(a)

    a_out, c_out, u_out, info = tb01nd('N', 'U', a, c)

    assert info == 0

    # Compute eigenvalues after transformation
    eig_after = np.linalg.eigvals(a_out)

    # Eigenvalues should be preserved (sort for comparison)
    eig_before_sorted = np.sort(eig_before)
    eig_after_sorted = np.sort(eig_after)

    assert_allclose(eig_before_sorted.real, eig_after_sorted.real, rtol=1e-13, atol=1e-14)
    assert_allclose(eig_before_sorted.imag, eig_after_sorted.imag, rtol=1e-13, atol=1e-14)


def test_tb01nd_p_eq_n():
    """
    Test case where P == N (C becomes square upper triangular, A stays full).

    Random seed: 888 (for reproducibility)
    """
    from slicot import tb01nd

    np.random.seed(888)
    n, p = 4, 4  # p == n

    a = np.random.randn(n, n).astype(float, order='F')
    c = np.random.randn(p, n).astype(float, order='F')
    a_orig = a.copy()
    c_orig = c.copy()

    a_out, c_out, u_out, info = tb01nd('I', 'U', a, c)

    assert info == 0

    # Verify U is orthogonal
    assert_allclose(u_out.T @ u_out, np.eye(n), rtol=1e-14, atol=1e-14)

    # Verify similarity transformation
    assert_allclose(a_out, u_out.T @ a_orig @ u_out, rtol=1e-14, atol=1e-14)
    assert_allclose(c_out, c_orig @ u_out, rtol=1e-14, atol=1e-14)


def test_tb01nd_n_zero():
    """Test edge case with N=0 and P=0."""
    from slicot import tb01nd

    a = np.empty((0, 0), order='F', dtype=float)
    c = np.empty((0, 0), order='F', dtype=float)  # P must be <= N, so P=0 when N=0

    a_out, c_out, u_out, info = tb01nd('I', 'U', a, c)

    assert info == 0
    assert a_out.shape == (0, 0)
    assert c_out.shape == (0, 0)


def test_tb01nd_p_zero():
    """Test edge case with P=0 - A should be unchanged."""
    from slicot import tb01nd

    np.random.seed(111)
    n = 4

    a = np.random.randn(n, n).astype(float, order='F')
    c = np.empty((0, n), order='F', dtype=float)
    a_orig = a.copy()

    a_out, c_out, u_out, info = tb01nd('I', 'U', a, c)

    assert info == 0
    # When P=0, A is unchanged
    assert_allclose(a_out, a_orig, rtol=1e-14, atol=1e-14)
    # U should be identity when no transformation needed
    assert_allclose(u_out, np.eye(n), rtol=1e-14, atol=1e-14)


def test_tb01nd_invalid_jobu():
    """Test error handling for invalid JOBU parameter."""
    from slicot import tb01nd

    a = np.array([[1.0, 2.0], [3.0, 4.0]], order='F', dtype=float)
    c = np.array([[1.0, 2.0]], order='F', dtype=float)

    with pytest.raises(ValueError):
        tb01nd('X', 'U', a, c)


def test_tb01nd_invalid_uplo():
    """Test error handling for invalid UPLO parameter."""
    from slicot import tb01nd

    a = np.array([[1.0, 2.0], [3.0, 4.0]], order='F', dtype=float)
    c = np.array([[1.0, 2.0]], order='F', dtype=float)

    with pytest.raises(ValueError):
        tb01nd('N', 'X', a, c)


def test_tb01nd_p_greater_than_n():
    """Test error handling when P > N (invalid)."""
    from slicot import tb01nd

    # P cannot exceed N per Fortran docs: 0 <= P <= N
    # However, looking at HTML doc this seems valid (trapezoidal C)
    # Let's check by just verifying it returns proper info
    # Actually, the Fortran has: IF( P.LT.0 .OR. P.GT.N )
    # So P > N should give INFO = -4

    n, p = 3, 5
    a = np.random.randn(n, n).astype(float, order='F')
    c = np.random.randn(p, n).astype(float, order='F')

    a_out, c_out, u_out, info = tb01nd('N', 'U', a, c)

    # Fortran docs: 0 <= P <= N, so P > N should give error
    assert info == -4



"""
Tests for TB01IZ: Balancing a system matrix for complex state-space systems.

TB01IZ reduces the 1-norm of the system matrix S = [[A, B], [C, 0]] by applying
diagonal similarity transformations to balance rows and columns.

Reference: SLICOT-Reference/doc/TB01IZ.html
"""

import numpy as np
import pytest
from slicot import tb01iz


"""Basic functionality tests using HTML doc example."""

def test_html_doc_example():
    """
    Test using the exact example from SLICOT HTML documentation.

    N=5, M=2, P=5, JOB='A', MAXRED=0.0
    Validates: eigenvalue preservation, reconstruction, norm reduction.
    Note: Balancing is iterative and may converge to different (but valid) solutions.
    """
    n, m, p = 5, 2, 5

    a = np.array([
        [0.0+0.0j, 1.0e3+0.0j, 0.0+0.0j, 0.0+0.0j, 0.0+0.0j],
        [-1.58e6+0.0j, -1.257e3+0.0j, 0.0+0.0j, 0.0+0.0j, 0.0+0.0j],
        [3.541e14+0.0j, 0.0+0.0j, -1.434e3+0.0j, 0.0+0.0j, -5.33e11+0.0j],
        [0.0+0.0j, 0.0+0.0j, 0.0+0.0j, 0.0+0.0j, 1.0+0.0j],
        [0.0+0.0j, 0.0+0.0j, 0.0+0.0j, -1.863e4+0.0j, -1.482+0.0j]
    ], order='F', dtype=complex)

    b = np.array([
        [0.0+0.0j, 0.0+0.0j],
        [1.103e2+0.0j, 0.0+0.0j],
        [0.0+0.0j, 0.0+0.0j],
        [0.0+0.0j, 0.0+0.0j],
        [0.0+0.0j, 8.333e-3+0.0j]
    ], order='F', dtype=complex)

    c = np.array([
        [1.0+0.0j, 0.0+0.0j, 0.0+0.0j, 0.0+0.0j, 0.0+0.0j],
        [0.0+0.0j, 0.0+0.0j, 1.0+0.0j, 0.0+0.0j, 0.0+0.0j],
        [0.0+0.0j, 0.0+0.0j, 0.0+0.0j, 1.0+0.0j, 0.0+0.0j],
        [6.664e-1+0.0j, 0.0+0.0j, -6.2e-13+0.0j, 0.0+0.0j, 0.0+0.0j],
        [0.0+0.0j, 0.0+0.0j, -1.0e-3+0.0j, 1.896e6+0.0j, 1.508e2+0.0j]
    ], order='F', dtype=complex)

    a_orig, b_orig, c_orig = a.copy(), b.copy(), c.copy()

    a_bal, b_bal, c_bal, maxred_out, scale, info = tb01iz('A', a, b, c, 0.0)

    assert info == 0
    assert maxred_out > 1.0

    d = np.diag(scale)
    d_inv = np.diag(1.0 / scale)
    a_reconstructed = d @ a_bal @ d_inv
    b_reconstructed = d @ b_bal
    c_reconstructed = c_bal @ d_inv

    np.testing.assert_allclose(a_reconstructed, a_orig, rtol=1e-12, atol=1e-14)
    np.testing.assert_allclose(b_reconstructed, b_orig, rtol=1e-12, atol=1e-14)
    np.testing.assert_allclose(c_reconstructed, c_orig, rtol=1e-12, atol=1e-14)

    eig_orig = np.linalg.eigvals(a_orig)
    eig_bal = np.linalg.eigvals(a_bal)
    np.testing.assert_allclose(
        sorted(np.abs(eig_orig)),
        sorted(np.abs(eig_bal)),
        rtol=1e-10
    )

    assert all(np.log10(s) == int(np.log10(s)) for s in scale if s != 0)


"""Test mathematical properties of the balancing transformation."""

def test_similarity_transformation_preserves_eigenvalues():
    """
    Validate that diagonal similarity transformation preserves eigenvalues.

    The balanced matrix inv(D)*A*D should have the same eigenvalues as A.
    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    n, m, p = 4, 2, 3

    a = (np.random.randn(n, n) + 1j * np.random.randn(n, n)).astype(complex, order='F')
    b = (np.random.randn(n, m) + 1j * np.random.randn(n, m)).astype(complex, order='F')
    c = (np.random.randn(p, n) + 1j * np.random.randn(p, n)).astype(complex, order='F')

    eig_before = np.linalg.eigvals(a)

    a_bal, b_bal, c_bal, maxred_out, scale, info = tb01iz('A', a.copy(), b.copy(), c.copy(), 0.0)
    assert info == 0

    eig_after = np.linalg.eigvals(a_bal)

    np.testing.assert_allclose(
        sorted(np.abs(eig_before)),
        sorted(np.abs(eig_after)),
        rtol=1e-10
    )

def test_reconstruction_from_scale():
    """
    Validate that we can reconstruct the original system from balanced one using scale.

    A = D * A_bal * inv(D)
    B = D * B_bal
    C = C_bal * inv(D)

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)
    n, m, p = 3, 2, 2

    a = (np.random.randn(n, n) + 1j * np.random.randn(n, n)).astype(complex, order='F')
    b = (np.random.randn(n, m) + 1j * np.random.randn(n, m)).astype(complex, order='F')
    c = (np.random.randn(p, n) + 1j * np.random.randn(p, n)).astype(complex, order='F')

    a_orig = a.copy()
    b_orig = b.copy()
    c_orig = c.copy()

    a_bal, b_bal, c_bal, maxred_out, scale, info = tb01iz('A', a.copy(), b.copy(), c.copy(), 0.0)
    assert info == 0

    d = np.diag(scale)
    d_inv = np.diag(1.0 / scale)

    a_reconstructed = d @ a_bal @ d_inv
    b_reconstructed = d @ b_bal
    c_reconstructed = c_bal @ d_inv

    np.testing.assert_allclose(a_reconstructed, a_orig, rtol=1e-12, atol=1e-14)
    np.testing.assert_allclose(b_reconstructed, b_orig, rtol=1e-12, atol=1e-14)
    np.testing.assert_allclose(c_reconstructed, c_orig, rtol=1e-12, atol=1e-14)

def test_norm_reduction():
    """
    Validate that balancing reduces or maintains system matrix 1-norm.

    Random seed: 456 (for reproducibility)
    """
    np.random.seed(456)
    n, m, p = 5, 2, 3

    scale_factors = 10.0 ** np.random.uniform(-5, 5, n)
    a = np.diag(scale_factors) @ np.random.randn(n, n) @ np.diag(1.0 / scale_factors)
    a = (a + 1j * np.random.randn(n, n) * 0.1).astype(complex, order='F')
    b = (np.random.randn(n, m) + 1j * np.random.randn(n, m)).astype(complex, order='F')
    c = (np.random.randn(p, n) + 1j * np.random.randn(p, n)).astype(complex, order='F')

    def compute_system_norm(a, b, c):
        s_top = np.hstack([a, b])
        s_bottom = np.hstack([c, np.zeros((c.shape[0], b.shape[1]), dtype=complex)])
        s = np.vstack([s_top, s_bottom])
        return np.max(np.sum(np.abs(s), axis=0))

    norm_before = compute_system_norm(a, b, c)

    a_bal, b_bal, c_bal, maxred_out, scale, info = tb01iz('A', a.copy(), b.copy(), c.copy(), 0.0)
    assert info == 0

    norm_after = compute_system_norm(a_bal, b_bal, c_bal)

    assert norm_after <= norm_before * 1.1


"""Test different JOB parameter modes."""

def test_job_a_balances_all():
    """
    Test JOB='A': All matrices (A, B, C) are involved in balancing.

    Random seed: 789 (for reproducibility)
    """
    np.random.seed(789)
    n, m, p = 3, 2, 2

    a = (np.random.randn(n, n) + 1j * np.random.randn(n, n)).astype(complex, order='F')
    b = (np.random.randn(n, m) + 1j * np.random.randn(n, m)).astype(complex, order='F')
    c = (np.random.randn(p, n) + 1j * np.random.randn(p, n)).astype(complex, order='F')

    a_orig, b_orig, c_orig = a.copy(), b.copy(), c.copy()

    a_bal, b_bal, c_bal, maxred, scale, info = tb01iz('A', a, b, c, 0.0)

    assert info == 0
    assert scale.shape == (n,)

def test_job_b_balances_a_and_b_only():
    """
    Test JOB='B': Only A and B matrices are involved in balancing.
    C is not used in computing the scaling.

    Random seed: 111 (for reproducibility)
    """
    np.random.seed(111)
    n, m, p = 3, 2, 2

    a = (np.random.randn(n, n) + 1j * np.random.randn(n, n)).astype(complex, order='F')
    b = (np.random.randn(n, m) + 1j * np.random.randn(n, m)).astype(complex, order='F')
    c = (np.random.randn(p, n) + 1j * np.random.randn(p, n)).astype(complex, order='F')

    a_bal, b_bal, c_bal, maxred, scale, info = tb01iz('B', a.copy(), b.copy(), c.copy(), 0.0)

    assert info == 0

def test_job_c_balances_a_and_c_only():
    """
    Test JOB='C': Only A and C matrices are involved in balancing.
    B is not used in computing the scaling.

    Random seed: 222 (for reproducibility)
    """
    np.random.seed(222)
    n, m, p = 3, 2, 2

    a = (np.random.randn(n, n) + 1j * np.random.randn(n, n)).astype(complex, order='F')
    b = (np.random.randn(n, m) + 1j * np.random.randn(n, m)).astype(complex, order='F')
    c = (np.random.randn(p, n) + 1j * np.random.randn(p, n)).astype(complex, order='F')

    a_bal, b_bal, c_bal, maxred, scale, info = tb01iz('C', a.copy(), b.copy(), c.copy(), 0.0)

    assert info == 0

def test_job_n_balances_a_only():
    """
    Test JOB='N': Only A matrix is balanced (B and C not used).

    Random seed: 333 (for reproducibility)
    """
    np.random.seed(333)
    n, m, p = 3, 2, 2

    a = (np.random.randn(n, n) + 1j * np.random.randn(n, n)).astype(complex, order='F')
    b = (np.random.randn(n, m) + 1j * np.random.randn(n, m)).astype(complex, order='F')
    c = (np.random.randn(p, n) + 1j * np.random.randn(p, n)).astype(complex, order='F')

    a_bal, b_bal, c_bal, maxred, scale, info = tb01iz('N', a.copy(), b.copy(), c.copy(), 0.0)

    assert info == 0


"""Edge cases and boundary conditions."""

def test_n_equals_zero():
    """Test with n=0 (trivial case)."""
    a = np.array([], dtype=complex, order='F').reshape(0, 0)
    b = np.array([], dtype=complex, order='F').reshape(0, 1)
    c = np.array([], dtype=complex, order='F').reshape(1, 0)

    a_bal, b_bal, c_bal, maxred, scale, info = tb01iz('A', a, b, c, 0.0)
    assert info == 0

def test_identity_system():
    """Test with identity A matrix (already balanced)."""
    n, m, p = 3, 1, 1
    a = np.eye(n, dtype=complex, order='F')
    b = np.ones((n, m), dtype=complex, order='F')
    c = np.ones((p, n), dtype=complex, order='F')

    a_bal, b_bal, c_bal, maxred, scale, info = tb01iz('A', a, b, c, 0.0)

    assert info == 0
    np.testing.assert_allclose(scale, np.ones(n), rtol=1e-10)

def test_zero_matrix():
    """Test with zero system matrix."""
    n, m, p = 3, 2, 2
    a = np.zeros((n, n), dtype=complex, order='F')
    b = np.zeros((n, m), dtype=complex, order='F')
    c = np.zeros((p, n), dtype=complex, order='F')

    a_bal, b_bal, c_bal, maxred, scale, info = tb01iz('A', a, b, c, 0.0)

    assert info == 0

def test_m_equals_zero():
    """Test with m=0 (no inputs)."""
    n, m, p = 3, 0, 2
    a = np.eye(n, dtype=complex, order='F')
    b = np.array([], dtype=complex, order='F').reshape(n, 0)
    c = np.ones((p, n), dtype=complex, order='F')

    a_bal, b_bal, c_bal, maxred, scale, info = tb01iz('A', a, b, c, 0.0)
    assert info == 0

def test_p_equals_zero():
    """Test with p=0 (no outputs)."""
    n, m, p = 3, 2, 0
    a = np.eye(n, dtype=complex, order='F')
    b = np.ones((n, m), dtype=complex, order='F')
    c = np.array([], dtype=complex, order='F').reshape(0, n)

    a_bal, b_bal, c_bal, maxred, scale, info = tb01iz('A', a, b, c, 0.0)
    assert info == 0


"""Tests with complex values."""

def test_purely_imaginary_matrix():
    """
    Test with purely imaginary system matrices.

    Random seed: 444 (for reproducibility)
    """
    np.random.seed(444)
    n, m, p = 3, 2, 2

    a = 1j * np.random.randn(n, n).astype(complex, order='F')
    b = 1j * np.random.randn(n, m).astype(complex, order='F')
    c = 1j * np.random.randn(p, n).astype(complex, order='F')

    a_orig = a.copy()

    a_bal, b_bal, c_bal, maxred, scale, info = tb01iz('A', a.copy(), b.copy(), c.copy(), 0.0)

    assert info == 0

    eig_before = np.linalg.eigvals(a_orig)
    eig_after = np.linalg.eigvals(a_bal)
    np.testing.assert_allclose(
        sorted(np.abs(eig_before)),
        sorted(np.abs(eig_after)),
        rtol=1e-10
    )

def test_general_complex_matrix():
    """
    Test with general complex values (real + imaginary parts).

    Random seed: 555 (for reproducibility)
    """
    np.random.seed(555)
    n, m, p = 4, 2, 3

    a = (np.random.randn(n, n) + 1j * np.random.randn(n, n)).astype(complex, order='F')
    b = (np.random.randn(n, m) + 1j * np.random.randn(n, m)).astype(complex, order='F')
    c = (np.random.randn(p, n) + 1j * np.random.randn(p, n)).astype(complex, order='F')

    a_orig, b_orig, c_orig = a.copy(), b.copy(), c.copy()

    a_bal, b_bal, c_bal, maxred, scale, info = tb01iz('A', a.copy(), b.copy(), c.copy(), 0.0)

    assert info == 0

    d = np.diag(scale)
    d_inv = np.diag(1.0 / scale)

    a_reconstructed = d @ a_bal @ d_inv
    np.testing.assert_allclose(a_reconstructed, a_orig, rtol=1e-12, atol=1e-14)


"""Test error handling."""

def test_invalid_job_parameter():
    """Test invalid JOB parameter returns error."""
    n, m, p = 2, 1, 1
    a = np.eye(n, dtype=complex, order='F')
    b = np.ones((n, m), dtype=complex, order='F')
    c = np.ones((p, n), dtype=complex, order='F')

    a_bal, b_bal, c_bal, maxred, scale, info = tb01iz('X', a, b, c, 0.0)
    assert info == -1

def test_invalid_maxred_value():
    """Test MAXRED between 0 and 1 returns error."""
    n, m, p = 2, 1, 1
    a = np.eye(n, dtype=complex, order='F')
    b = np.ones((n, m), dtype=complex, order='F')
    c = np.ones((p, n), dtype=complex, order='F')

    a_bal, b_bal, c_bal, maxred, scale, info = tb01iz('A', a, b, c, 0.5)
    assert info == -5


"""Tests for MAXRED parameter behavior."""

def test_maxred_zero_uses_default():
    """Test MAXRED=0 uses default value of 10."""
    np.random.seed(666)
    n, m, p = 3, 2, 2

    a = (np.random.randn(n, n) + 1j * np.random.randn(n, n)).astype(complex, order='F')
    b = (np.random.randn(n, m) + 1j * np.random.randn(n, m)).astype(complex, order='F')
    c = (np.random.randn(p, n) + 1j * np.random.randn(p, n)).astype(complex, order='F')

    a_bal, b_bal, c_bal, maxred_out, scale, info = tb01iz('A', a, b, c, 0.0)

    assert info == 0

def test_maxred_positive_custom_value():
    """Test custom MAXRED > 1 value."""
    np.random.seed(777)
    n, m, p = 3, 2, 2

    a = (np.random.randn(n, n) + 1j * np.random.randn(n, n)).astype(complex, order='F')
    b = (np.random.randn(n, m) + 1j * np.random.randn(n, m)).astype(complex, order='F')
    c = (np.random.randn(p, n) + 1j * np.random.randn(p, n)).astype(complex, order='F')

    a_bal, b_bal, c_bal, maxred_out, scale, info = tb01iz('A', a, b, c, 5.0)

    assert info == 0



import unittest
import numpy as np
from slicot import mb04ow

def test_mb04ow_basic():
    """
    Validate MB04OW QR update using Gram matrix preservation
    
    MB04OW updates the QR factorization of:
         ( U  )   = Q * ( R )
         ( x' )         ( 0 )
    
    where U = [ U1 U2 ]
              [ 0  T  ]
    
    It also updates [B; C; d'] by applying Q'.
    """
    # Reproducibility
    np.random.seed(42)
    
    m, n, p = 4, 3, 2
    lda = m + 2
    ldt = n + 2
    ldb = m + 2
    ldc = n + 2
    
    # Generate random matrices (Fortran order)
    # U1: mxm upper triangular
    # U2: mxn full
    # T: nxn upper triangular
    u1 = np.triu(np.random.rand(m, m))
    u2 = np.random.rand(m, n)
    t = np.triu(np.random.rand(n, n))
    
    # Store U1, U2 in A
    a = np.zeros((lda, n + m), order='F')
    a[:m, :m] = u1
    a[:m, m:m+n] = u2
    
    # Store T in T_arr
    t_arr = np.zeros((ldt, n), order='F')
    t_arr[:n, :n] = t
    
    # B, C
    b = np.random.rand(m, p)
    b_arr = np.zeros((ldb, p), order='F')
    b_arr[:m, :p] = b
    
    c = np.random.rand(n, p)
    c_arr = np.zeros((ldc, p), order='F')
    c_arr[:n, :p] = c
    
    # x (m+n vector)
    x = np.random.rand(m + n)
    x_arr = x.copy() # MB04OW modifies X
    
    # d (p vector)
    d = np.random.rand(p)
    d_arr = d.copy() # MB04OW modifies D
    
    # Construct the big matrix K for verification
    # K = [ U1  U2  B ]
    #     [ 0   T   C ]
    #     [ x'      d']
    #
    # But carefully map x: x is length m+n
    # x[0:m] corresponds to U1 columns
    # x[m:m+n] corresponds to U2/T columns
    
    # Dimensions of K: (m + n + 1) x (m + n + p)
    K = np.zeros((m + n + 1, m + n + p))
    
    # Row 0..m-1
    K[:m, :m] = u1
    K[:m, m:m+n] = u2
    K[:m, m+n:] = b
    
    # Row m..m+n-1
    K[m:m+n, m:m+n] = t
    K[m:m+n, m+n:] = c
    
    # Row m+n (the update row)
    K[m+n, :m+n] = x
    K[m+n, m+n:] = d
    
    # Call MB04OW
    # mb04ow(m, n, p, a, t, x, b, c, d, incx=1, incd=1)
    # Note: python wrapper should handle array dimensions.
    # We assume the wrapper signature:
    # a_out, t_out, x_out, b_out, c_out, d_out = mb04ow(a, t, x, b, c, d)
    # OR inplace. Let's assume standard SLICOT wrapper pattern (return modified arrays).
    
    a_out, t_out, x_out, b_out, c_out, d_out = mb04ow(m, n, p, a, t_arr, x_arr, b_arr, c_arr, d_arr)
    
    # Reconstruct the updated matrix from outputs
    # The top m+n rows should contain the updated R and transformed B,C
    # The last row is implicitly zero in the first m+n cols, and contains transformed d in last p cols.
    
    # But wait, MB04OW doesn't return the full last row of the transformed matrix?
    # "The transformations performed are also applied to ... ( B' C' d )' "
    # Yes, d is updated.
    # x is updated (to zeroes? or contains information about rotations? 
    # "On exit, the content of X is changed." (Likely garbage or rotation info)
    # The documentation says "Q(j) being chosen to annihilate the jth element of x."
    # So mathematically, after rotations, x should be 0.
    # In practice, X is workspace.
    
    # Construct K_out from outputs
    K_out = np.zeros_like(K)
    
    # Extract R1, R2 from a_out
    r1 = a_out[:m, :m] # Upper triangular
    r2 = a_out[:m, m:m+n]
    
    # Extract R3 from t_out
    r3 = t_out[:n, :n] # Upper triangular
    
    # Extract B_out, C_out
    b_new = b_out[:m, :p]
    c_new = c_out[:n, :p]
    
    # Extract d_new
    d_new = d_out
    
    # Fill K_out
    K_out[:m, :m] = np.triu(r1)
    K_out[:m, m:m+n] = r2
    K_out[:m, m+n:] = b_new
    
    K_out[m:m+n, m:m+n] = np.triu(r3)
    K_out[m:m+n, m+n:] = c_new
    
    K_out[m+n, m+n:] = d_new
    
    # Now, K_out should be Q^T * K for some orthogonal Q.
    # Thus K_out^T * K_out should equal K^T * K.
    
    gram_in = K.T @ K
    gram_out = K_out.T @ K_out
    
    np.testing.assert_allclose(gram_out, gram_in, rtol=1e-12, atol=1e-14)
    
    # Also check that R1, R3 are upper triangular
    np.testing.assert_allclose(np.tril(r1, -1), 0, atol=1e-15)
    np.testing.assert_allclose(np.tril(r3, -1), 0, atol=1e-15)

def test_mb04ow_orthogonality():
    """
    Validate Q^T @ Q = I for orthogonal transformation.

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)

    m, n, p = 3, 2, 2
    lda = m + 1
    ldt = n + 1
    ldb = m + 1
    ldc = n + 1

    u1 = np.triu(np.random.rand(m, m))
    u2 = np.random.rand(m, n)
    t = np.triu(np.random.rand(n, n))

    a = np.zeros((lda, n + m), order='F')
    a[:m, :m] = u1
    a[:m, m:m+n] = u2

    t_arr = np.zeros((ldt, n), order='F')
    t_arr[:n, :n] = t

    b = np.random.rand(m, p)
    b_arr = np.zeros((ldb, p), order='F')
    b_arr[:m, :p] = b

    c = np.random.rand(n, p)
    c_arr = np.zeros((ldc, p), order='F')
    c_arr[:n, :p] = c

    x = np.random.rand(m + n)
    x_arr = x.copy()

    d = np.random.rand(p)
    d_arr = d.copy()

    K = np.zeros((m + n + 1, m + n + p))
    K[:m, :m] = u1
    K[:m, m:m+n] = u2
    K[:m, m+n:] = b
    K[m:m+n, m:m+n] = t
    K[m:m+n, m+n:] = c
    K[m+n, :m+n] = x
    K[m+n, m+n:] = d

    a_out, t_out, x_out, b_out, c_out, d_out = mb04ow(m, n, p, a, t_arr, x_arr, b_arr, c_arr, d_arr)

    # Reconstruct transformed matrix
    K_out = np.zeros_like(K)
    K_out[:m, :m] = np.triu(a_out[:m, :m])
    K_out[:m, m:m+n] = a_out[:m, m:m+n]
    K_out[:m, m+n:] = b_out[:m, :p]
    K_out[m:m+n, m:m+n] = np.triu(t_out[:n, :n])
    K_out[m:m+n, m+n:] = c_out[:n, :p]
    K_out[m+n, m+n:] = d_out

    # For orthogonal Q: K_out = Q^T @ K, so K = Q @ K_out
    # Compute Q by solving for it
    # Actually, verify orthogonality via Q^T @ Q = I
    # We can compute Q from relationship K = Q @ K_out
    # Use thin QR: Q @ R = K where R = K_out (upper part)

    # Alternative: verify column norms preserved
    for j in range(m + n + p):
        norm_before = np.linalg.norm(K[:, j])
        norm_after = np.linalg.norm(K_out[:, j])
        np.testing.assert_allclose(norm_after, norm_before, rtol=1e-12)

def test_mb04ow_edge_case_zero_dims():
    """
    Validate MB04OW with zero dimensions.

    Tests p=0 case (no extra columns to transform).
    """
    np.random.seed(456)

    m, n, p = 3, 2, 0
    lda = m
    ldt = n
    ldb = m
    ldc = n

    u1 = np.triu(np.random.rand(m, m))
    u2 = np.random.rand(m, n)
    t = np.triu(np.random.rand(n, n))

    a = np.zeros((lda, n + m), order='F')
    a[:m, :m] = u1
    a[:m, m:m+n] = u2

    t_arr = np.zeros((ldt, n), order='F')
    t_arr[:n, :n] = t

    b_arr = np.zeros((ldb, max(1, p)), order='F')
    c_arr = np.zeros((ldc, max(1, p)), order='F')

    x = np.random.rand(m + n)
    x_arr = x.copy()

    d_arr = np.zeros(max(1, p))

    a_out, t_out, x_out, b_out, c_out, d_out = mb04ow(m, n, p, a, t_arr, x_arr, b_arr, c_arr, d_arr)

    # Verify triangular structure preserved
    np.testing.assert_allclose(np.tril(a_out[:m, :m], -1), 0, atol=1e-15)
    np.testing.assert_allclose(np.tril(t_out[:n, :n], -1), 0, atol=1e-15)

if __name__ == '__main__':
    unittest.main()



"""
Tests for MB03WD: Periodic Schur Decomposition

MB03WD computes the Schur decomposition and eigenvalues of a product of matrices
H = H_1*H_2*...*H_p, with H_1 upper Hessenberg and H_2...H_p upper triangular,
without evaluating the product.
"""

import numpy as np
import pytest

from slicot import mb03vd, mb03wd


"""Basic functionality tests from SLICOT HTML documentation."""

def test_html_doc_example():
    """
    Test case from SLICOT HTML documentation.

    N=4, P=2, ILO=1, IHI=4, JOB='S', COMPZ='V'
    Two 4x4 matrices forming the product.

    The example first reduces to Hessenberg form using MB03VD,
    then applies MB03WD for periodic Schur decomposition.

    Expected eigenvalues (complex conjugate pairs and real):
        (6.449861, 7.817717)
        (6.449861, -7.817717)
        (0.091315, 0.0)
        (0.208964, 0.0)
    """
    n = 4
    p = 2
    ilo = 1
    ihi = 4
    iloz = 1
    ihiz = 4

    a = np.zeros((n, n, p), order='F', dtype=float)
    a[:, :, 0] = np.array([
        [1.5, -0.7, 3.5, -0.7],
        [1.0,  0.0, 2.0,  3.0],
        [1.5, -0.7, 2.5, -0.3],
        [1.0,  0.0, 2.0,  1.0]
    ], order='F', dtype=float)

    a[:, :, 1] = np.array([
        [1.5, -0.7, 3.5, -0.7],
        [1.0,  0.0, 2.0,  3.0],
        [1.5, -0.7, 2.5, -0.3],
        [1.0,  0.0, 2.0,  1.0]
    ], order='F', dtype=float)

    h, tau, info_vd = mb03vd(n, p, ilo, ihi, a)
    assert info_vd == 0

    h_result, z, wr, wi, info = mb03wd('S', 'I', n, p, ilo, ihi, iloz, ihiz, h)

    assert info == 0

    expected_wr = np.array([6.449861, 6.449861, 0.091315, 0.208964])
    expected_wi = np.array([7.817717, -7.817717, 0.0, 0.0])

    np.testing.assert_allclose(wr, expected_wr, rtol=1e-5, atol=1e-5)
    np.testing.assert_allclose(wi, expected_wi, rtol=1e-5, atol=1e-10)

    t1 = h_result[:, :, 0]
    t2 = h_result[:, :, 1]

    assert np.abs(t1[2, 0]) < 1e-10
    assert np.abs(t1[3, 0]) < 1e-10
    assert np.abs(t1[3, 1]) < 1e-10

    for i in range(n):
        for j in range(i):
            assert np.abs(t2[i, j]) < 1e-10, f"t2[{i},{j}] = {t2[i,j]} not zero"


"""Test eigenvalue-only computation (JOB='E')."""

def test_eigenvalues_only_mode():
    """
    Test JOB='E' mode which computes eigenvalues without full Schur form.

    Uses pre-reduced Hessenberg matrices (H_1 upper Hessenberg, H_2 upper triangular).
    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    n = 3
    p = 2
    ilo = 1
    ihi = n
    iloz = 1
    ihiz = n

    h1 = np.array([
        [2.0, 1.0, 0.5],
        [0.3, 1.5, 0.7],
        [0.0, 0.2, 1.0]
    ], order='F', dtype=float)

    h2 = np.triu(np.array([
        [1.0, 0.4, 0.2],
        [0.0, 0.8, 0.3],
        [0.0, 0.0, 0.5]
    ], order='F', dtype=float))

    h = np.zeros((n, n, p), order='F', dtype=float)
    h[:, :, 0] = h1
    h[:, :, 1] = h2

    product = h1 @ h2
    expected_eig = np.linalg.eigvals(product)
    expected_eig_sorted = np.sort(np.real(expected_eig))[::-1]

    h_copy = h.copy(order='F')

    h_result, z, wr, wi, info = mb03wd('E', 'N', n, p, ilo, ihi, iloz, ihiz, h_copy)

    assert info == 0
    assert wr.shape == (n,)
    assert wi.shape == (n,)

    computed_eig = []
    i = 0
    while i < n:
        if wi[i] == 0.0:
            computed_eig.append(wr[i])
            i += 1
        else:
            computed_eig.append(complex(wr[i], wi[i]))
            computed_eig.append(complex(wr[i+1], wi[i+1]))
            i += 2
    computed_eig_sorted = np.sort(np.real(computed_eig))[::-1]

    np.testing.assert_allclose(computed_eig_sorted, expected_eig_sorted, rtol=1e-3)


"""Mathematical property validation tests."""

def test_eigenvalue_preservation():
    """
    Validate eigenvalues of product are preserved by periodic Schur form.

    The eigenvalues of T_1*T_2*...*T_p should equal eigenvalues of H_1*H_2*...*H_p.

    Uses pre-reduced Hessenberg matrices (H_1 upper Hessenberg, H_2 upper triangular).
    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)
    n = 4
    p = 2
    ilo = 1
    ihi = n
    iloz = 1
    ihiz = n

    h1 = np.array([
        [1.2, 0.5, -0.3, 0.1],
        [0.8, -0.4, 0.7, 0.2],
        [0.0, 0.6, 1.1, -0.2],
        [0.0, 0.0, 0.9, 0.8]
    ], order='F', dtype=float)

    h2 = np.triu(np.array([
        [0.9, 0.2, -0.1, 0.3],
        [0.0, 1.1, 0.4, -0.2],
        [0.0, 0.0, 0.7, 0.5],
        [0.0, 0.0, 0.0, 1.3]
    ], order='F', dtype=float))

    h = np.zeros((n, n, p), order='F', dtype=float)
    h[:, :, 0] = h1.copy()
    h[:, :, 1] = h2.copy()

    product_before = h1 @ h2
    eig_before = np.linalg.eigvals(product_before)

    h_result, z, wr, wi, info = mb03wd('S', 'I', n, p, ilo, ihi, iloz, ihiz, h)

    assert info == 0

    t1 = h_result[:, :, 0]
    t2 = h_result[:, :, 1]
    product_after = t1 @ t2
    eig_after = np.linalg.eigvals(product_after)

    eig_before_sorted = sorted(eig_before, key=lambda x: (x.real, x.imag))
    eig_after_sorted = sorted(eig_after, key=lambda x: (x.real, x.imag))

    np.testing.assert_allclose(
        [e.real for e in eig_before_sorted],
        [e.real for e in eig_after_sorted],
        rtol=1e-3
    )
    np.testing.assert_allclose(
        [e.imag for e in eig_before_sorted],
        [e.imag for e in eig_after_sorted],
        rtol=1e-3,
        atol=1e-9
    )

def test_transformation_orthogonality():
    """
    Validate transformation matrices Z_i are orthogonal.

    Z_i' * Z_i = I for all i = 1, ..., p.

    Random seed: 456 (for reproducibility)
    """
    np.random.seed(456)
    n = 4
    p = 3
    ilo = 1
    ihi = n
    iloz = 1
    ihiz = n

    h = np.zeros((n, n, p), order='F', dtype=float)

    h[:, :, 0] = np.array([
        [2.0, 1.0, 0.5, 0.2],
        [0.5, 1.5, 0.8, 0.3],
        [0.0, 0.4, 1.2, 0.6],
        [0.0, 0.0, 0.3, 0.9]
    ], order='F', dtype=float)

    for k in range(1, p):
        h[:, :, k] = np.triu(np.random.randn(n, n)).astype(float, order='F')
        for i in range(n):
            h[i, i, k] = np.abs(h[i, i, k]) + 0.5

    h_result, z, wr, wi, info = mb03wd('S', 'I', n, p, ilo, ihi, iloz, ihiz, h)

    assert info == 0
    assert z.shape == (n, n, p)

    identity = np.eye(n, dtype=float)
    for k in range(p):
        z_k = z[:, :, k]
        orthogonality = z_k.T @ z_k
        np.testing.assert_allclose(orthogonality, identity, rtol=1e-12, atol=1e-12)

def test_transformation_property():
    """
    Validate periodic Schur transformation property.

    Z_1' * H_1 * Z_2 = T_1 (real Schur form)
    Z_j' * H_j * Z_{j+1} = T_j (upper triangular) for j >= 2
    with Z_{p+1} = Z_1.

    Uses pre-reduced Hessenberg matrices (H_1 upper Hessenberg, H_2 upper triangular).
    Random seed: 789 (for reproducibility)
    """
    np.random.seed(789)
    n = 4
    p = 2
    ilo = 1
    ihi = n
    iloz = 1
    ihiz = n

    h1_orig = np.array([
        [1.2, 0.5, -0.3, 0.1],
        [0.8, -0.4, 0.7, 0.2],
        [0.0, 0.6, 1.1, -0.2],
        [0.0, 0.0, 0.9, 0.8]
    ], order='F', dtype=float)

    h2_orig = np.triu(np.array([
        [0.9, 0.2, -0.1, 0.3],
        [0.0, 1.1, 0.4, -0.2],
        [0.0, 0.0, 0.7, 0.5],
        [0.0, 0.0, 0.0, 1.3]
    ], order='F', dtype=float))

    h = np.zeros((n, n, p), order='F', dtype=float)
    h[:, :, 0] = h1_orig.copy()
    h[:, :, 1] = h2_orig.copy()

    h_result, z, wr, wi, info = mb03wd('S', 'I', n, p, ilo, ihi, iloz, ihiz, h)

    assert info == 0

    z1 = z[:, :, 0]
    z2 = z[:, :, 1]
    t1 = h_result[:, :, 0]
    t2 = h_result[:, :, 1]

    computed_t1 = z1.T @ h1_orig @ z2
    np.testing.assert_allclose(computed_t1, t1, rtol=1e-4, atol=1e-4)

    computed_t2 = z2.T @ h2_orig @ z1
    np.testing.assert_allclose(computed_t2, t2, rtol=1e-4, atol=1e-4)


"""Edge case tests."""

def test_single_element():
    """Test with n=1 (single element matrices)."""
    n = 1
    p = 2
    ilo = 1
    ihi = 1
    iloz = 1
    ihiz = 1

    h = np.zeros((n, n, p), order='F', dtype=float)
    h[0, 0, 0] = 2.0
    h[0, 0, 1] = 3.0

    h_result, z, wr, wi, info = mb03wd('S', 'I', n, p, ilo, ihi, iloz, ihiz, h)

    assert info == 0
    np.testing.assert_allclose(wr[0], 6.0, rtol=1e-14)
    np.testing.assert_allclose(wi[0], 0.0, atol=1e-14)

def test_single_matrix_p1():
    """Test with p=1 (single matrix, reduces to standard Schur)."""
    n = 3
    p = 1
    ilo = 1
    ihi = n
    iloz = 1
    ihiz = n

    h = np.zeros((n, n, p), order='F', dtype=float)
    h[:, :, 0] = np.array([
        [4.0, 1.0, 2.0],
        [1.0, 3.0, 1.0],
        [0.0, 1.0, 2.0]
    ], order='F', dtype=float)

    h_result, z, wr, wi, info = mb03wd('S', 'I', n, p, ilo, ihi, iloz, ihiz, h)

    assert info == 0

    expected_eig = np.linalg.eigvals(h[:, :, 0])
    expected_eig_real = np.sort(np.real(expected_eig))[::-1]

    computed_real = []
    i = 0
    while i < n:
        if wi[i] == 0.0:
            computed_real.append(wr[i])
            i += 1
        else:
            computed_real.append(wr[i])
            computed_real.append(wr[i+1])
            i += 2
    computed_real_sorted = np.sort(computed_real)[::-1]

    np.testing.assert_allclose(computed_real_sorted, expected_eig_real, rtol=1e-10)

def test_compz_v_accumulate():
    """
    Test COMPZ='V' mode which accumulates transformations.

    Random seed: 999 (for reproducibility)
    """
    np.random.seed(999)
    n = 3
    p = 2
    ilo = 1
    ihi = n
    iloz = 1
    ihiz = n

    h = np.zeros((n, n, p), order='F', dtype=float)
    h[:, :, 0] = np.array([
        [2.0, 1.0, 0.5],
        [0.3, 1.5, 0.7],
        [0.0, 0.2, 1.0]
    ], order='F', dtype=float)
    h[:, :, 1] = np.triu(np.array([
        [1.0, 0.4, 0.2],
        [0.0, 0.8, 0.3],
        [0.0, 0.0, 0.5]
    ], order='F', dtype=float))

    z = np.zeros((n, n, p), order='F', dtype=float)
    for k in range(p):
        z[:, :, k] = np.eye(n, dtype=float, order='F')

    h_result, z_result, wr, wi, info = mb03wd('S', 'V', n, p, ilo, ihi, iloz, ihiz, h, z)

    assert info == 0

    for k in range(p):
        orthogonality = z_result[:, :, k].T @ z_result[:, :, k]
        np.testing.assert_allclose(orthogonality, np.eye(n), rtol=1e-12, atol=1e-12)


"""Error handling tests."""

def test_invalid_n():
    """Test with invalid n < 0."""
    n = -1
    p = 2
    ilo = 1
    ihi = 1
    iloz = 1
    ihiz = 1

    h = np.zeros((1, 1, p), order='F', dtype=float)

    with pytest.raises(ValueError):
        mb03wd('S', 'I', n, p, ilo, ihi, iloz, ihiz, h)

def test_invalid_p():
    """Test with invalid p < 1."""
    n = 3
    p = 0
    ilo = 1
    ihi = n
    iloz = 1
    ihiz = n

    h = np.zeros((n, n, 1), order='F', dtype=float)

    with pytest.raises(ValueError):
        mb03wd('S', 'I', n, p, ilo, ihi, iloz, ihiz, h)

def test_invalid_ilo_ihi():
    """Test with invalid ILO > IHI."""
    n = 4
    p = 2
    ilo = 3
    ihi = 2
    iloz = 1
    ihiz = n

    h = np.zeros((n, n, p), order='F', dtype=float)

    with pytest.raises(ValueError):
        mb03wd('S', 'I', n, p, ilo, ihi, iloz, ihiz, h)



"""
Tests for MC01SX - Compute variation of exponents in floating-point series.

Computes V = max(E(j)) - min(E(j)) for j where MANT(j) != 0.
This is used internally by SLICOT for polynomial scaling operations.
"""

import numpy as np
import pytest


def test_mc01sx_basic():
    """
    Test basic exponent variation calculation.

    Exponents: [5, 10, 3, 8]
    Mantissas: [1.0, 2.0, 3.0, 4.0] (all non-zero)
    Expected variation: 10 - 3 = 7
    """
    from slicot import mc01sx

    e = np.array([5, 10, 3, 8], dtype=np.int32)
    mant = np.array([1.0, 2.0, 3.0, 4.0], dtype=np.float64)

    v = mc01sx(e, mant)

    assert v == 7


def test_mc01sx_with_zero_mantissa():
    """
    Test that zero mantissas are ignored.

    Exponents: [5, 100, 3, 8]  (100 should be ignored)
    Mantissas: [1.0, 0.0, 3.0, 4.0]
    Expected variation: 8 - 3 = 5 (ignoring the 100 at index 1)
    """
    from slicot import mc01sx

    e = np.array([5, 100, 3, 8], dtype=np.int32)
    mant = np.array([1.0, 0.0, 3.0, 4.0], dtype=np.float64)

    v = mc01sx(e, mant)

    assert v == 5


def test_mc01sx_single_element():
    """
    Test with single element.

    Variation should be 0.
    """
    from slicot import mc01sx

    e = np.array([42], dtype=np.int32)
    mant = np.array([1.5], dtype=np.float64)

    v = mc01sx(e, mant)

    assert v == 0


def test_mc01sx_all_same_exponent():
    """
    Test when all non-zero entries have the same exponent.

    Variation should be 0.
    """
    from slicot import mc01sx

    e = np.array([5, 5, 5, 5], dtype=np.int32)
    mant = np.array([1.0, 2.0, 3.0, 4.0], dtype=np.float64)

    v = mc01sx(e, mant)

    assert v == 0


def test_mc01sx_negative_exponents():
    """
    Test with negative exponents.

    Exponents: [-10, -5, -15, -3]
    Mantissas: [1.0, 1.0, 1.0, 1.0]
    Expected variation: -3 - (-15) = 12
    """
    from slicot import mc01sx

    e = np.array([-10, -5, -15, -3], dtype=np.int32)
    mant = np.array([1.0, 1.0, 1.0, 1.0], dtype=np.float64)

    v = mc01sx(e, mant)

    assert v == 12


def test_mc01sx_mixed_sign_exponents():
    """
    Test with mixed positive and negative exponents.

    Exponents: [-5, 0, 10, -10]
    Mantissas: [1.0, 1.0, 1.0, 1.0]
    Expected variation: 10 - (-10) = 20
    """
    from slicot import mc01sx

    e = np.array([-5, 0, 10, -10], dtype=np.int32)
    mant = np.array([1.0, 1.0, 1.0, 1.0], dtype=np.float64)

    v = mc01sx(e, mant)

    assert v == 20


def test_mc01sx_first_only_nonzero():
    """
    Test when only first mantissa is non-zero.

    Variation should be 0.
    """
    from slicot import mc01sx

    e = np.array([5, 100, 200, -50], dtype=np.int32)
    mant = np.array([1.0, 0.0, 0.0, 0.0], dtype=np.float64)

    v = mc01sx(e, mant)

    assert v == 0



"""
Tests for SB10YD: Frequency response fitting with stable minimum phase system.

SB10YD fits frequency response data with a SISO state-space system (A, B, C, D).
"""

import unittest
import numpy as np
from slicot import sb10yd


class TestSB10YD(unittest.TestCase):

    def test_sb10yd_continuous_first_order(self):
        """
        Test SB10YD with continuous-time first order system H(s) = 1/(s+1).

        The frequency response is H(jw) = 1/(1+jw).
        Random seed: N/A (deterministic test data from analytical formula)
        """
        n_points = 100
        omega = np.linspace(0.1, 10.0, n_points)
        rfrdat = np.zeros(n_points, dtype=float)
        ifrdat = np.zeros(n_points, dtype=float)

        for i in range(n_points):
            w = omega[i]
            denom = 1.0 + w * w
            rfrdat[i] = 1.0 / denom
            ifrdat[i] = -w / denom

        n = 1
        discfl = 0  # Continuous
        flag = 1    # Stable minimum phase
        tol = 0.0

        A, B, C, D, N_out, info = sb10yd(discfl, flag, n, rfrdat, ifrdat, omega, tol)

        self.assertEqual(info, 0)
        self.assertEqual(N_out, 1)

        # Eigenvalues should be stable (negative real part)
        eig_a = np.linalg.eigvals(A)
        self.assertTrue(np.all(np.real(eig_a) < 0), "A should be stable (negative eigenvalues)")

        # Verify frequency response at test points
        for w_test in [0.5, 1.0, 2.0]:
            s = 1j * w_test
            resolvent = np.linalg.inv(s * np.eye(N_out) - A)
            h_val = C @ resolvent @ B + D
            expected = 1.0 / (1.0 + 1j * w_test)
            np.testing.assert_allclose(np.abs(h_val), np.abs(expected), rtol=0.15)

    def test_sb10yd_discrete_first_order(self):
        """
        Test SB10YD with discrete-time first order system H(z) = z/(z-0.5).

        Random seed: N/A (deterministic test data from analytical formula)
        """
        n_points = 100
        omega = np.linspace(0.01, 3.0, n_points)  # Must be < pi for discrete
        rfrdat = np.zeros(n_points, dtype=float)
        ifrdat = np.zeros(n_points, dtype=float)

        for i in range(n_points):
            z = np.exp(1j * omega[i])
            h = z / (z - 0.5)
            rfrdat[i] = np.real(h)
            ifrdat[i] = np.imag(h)

        n = 1
        discfl = 1  # Discrete
        flag = 1    # Stable minimum phase
        tol = 0.0

        A, B, C, D, N_out, info = sb10yd(discfl, flag, n, rfrdat, ifrdat, omega, tol)

        self.assertEqual(info, 0)

        # Eigenvalues should be inside unit disk (modulus < 1)
        eig_a = np.linalg.eigvals(A)
        self.assertTrue(np.all(np.abs(eig_a) < 1.0), "A should be stable (eigenvalues inside unit disk)")

    def test_sb10yd_second_order_system(self):
        """
        Test SB10YD with second order system: H(s) = 1/((s+1)(s+2)).

        Random seed: N/A (deterministic test data from analytical formula)
        """
        n_points = 150
        omega = np.linspace(0.05, 15.0, n_points)
        rfrdat = np.zeros(n_points, dtype=float)
        ifrdat = np.zeros(n_points, dtype=float)

        for i in range(n_points):
            w = omega[i]
            s = 1j * w
            h = 1.0 / ((s + 1.0) * (s + 2.0))
            rfrdat[i] = np.real(h)
            ifrdat[i] = np.imag(h)

        n = 2
        discfl = 0  # Continuous
        flag = 1    # Stable minimum phase
        tol = 0.0

        A, B, C, D, N_out, info = sb10yd(discfl, flag, n, rfrdat, ifrdat, omega, tol)

        self.assertEqual(info, 0)
        self.assertEqual(N_out, 2)

        # Eigenvalues should be stable
        eig_a = np.linalg.eigvals(A)
        self.assertTrue(np.all(np.real(eig_a) < 0), "A should be stable")

        # Verify DC gain: H(0) = 1/(1*2) = 0.5
        resolvent = np.linalg.inv(-A)
        h_dc = C @ resolvent @ B + D
        np.testing.assert_allclose(np.real(h_dc), 0.5, rtol=0.15)

    def test_sb10yd_no_stability_constraint(self):
        """
        Test SB10YD without stability constraint (FLAG=0).

        Random seed: N/A (deterministic test data)
        """
        n_points = 100
        omega = np.linspace(0.1, 10.0, n_points)
        rfrdat = np.zeros(n_points, dtype=float)
        ifrdat = np.zeros(n_points, dtype=float)

        for i in range(n_points):
            w = omega[i]
            denom = 1.0 + w * w
            rfrdat[i] = 1.0 / denom
            ifrdat[i] = -w / denom

        n = 1
        discfl = 0
        flag = 0  # No stability constraint
        tol = 0.0

        A, B, C, D, N_out, info = sb10yd(discfl, flag, n, rfrdat, ifrdat, omega, tol)

        self.assertEqual(info, 0)
        self.assertEqual(N_out, 1)

    def test_sb10yd_zero_order(self):
        """
        Test SB10YD with N=0 (pure gain estimation).

        When N=0, the routine should estimate only D.
        """
        n_points = 50
        omega = np.linspace(0.1, 5.0, n_points)
        gain = 2.5
        rfrdat = np.ones(n_points, dtype=float) * gain
        ifrdat = np.zeros(n_points, dtype=float)

        n = 0
        discfl = 0  # Continuous
        flag = 0
        tol = 0.0

        A, B, C, D, N_out, info = sb10yd(discfl, flag, n, rfrdat, ifrdat, omega, tol)

        self.assertEqual(info, 0)
        self.assertEqual(N_out, 0)
        np.testing.assert_allclose(D, gain, rtol=1e-3)

    def test_sb10yd_freq_response_reconstruction(self):
        """
        Test that fitted system reproduces input frequency response.

        Mathematical property: H_fitted(jw) should approximate H_original(jw).
        Random seed: N/A (deterministic)
        """
        n_points = 100
        omega = np.linspace(0.1, 8.0, n_points)
        rfrdat = np.zeros(n_points, dtype=float)
        ifrdat = np.zeros(n_points, dtype=float)

        # Original system H(s) = 1/(s+1)
        for i in range(n_points):
            w = omega[i]
            denom = 1.0 + w * w
            rfrdat[i] = 1.0 / denom
            ifrdat[i] = -w / denom

        n = 1
        discfl = 0
        flag = 1
        tol = 0.0

        A, B, C, D, N_out, info = sb10yd(discfl, flag, n, rfrdat, ifrdat, omega, tol)
        self.assertEqual(info, 0)

        # Reconstruct frequency response
        max_rel_error = 0.0
        for i in range(0, n_points, 10):  # Check every 10th point
            w = omega[i]
            s = 1j * w
            resolvent = np.linalg.inv(s * np.eye(N_out) - A)
            h_fitted = C @ resolvent @ B + D
            h_original = rfrdat[i] + 1j * ifrdat[i]

            rel_error = np.abs(h_fitted - h_original) / np.abs(h_original)
            if rel_error > max_rel_error:
                max_rel_error = rel_error

        # Relative error should be reasonable for fitting (allowing for cepstrum method approximation)
        self.assertTrue(max_rel_error < 0.7, f"Max relative error {max_rel_error} too large")

    def test_sb10yd_minimum_phase(self):
        """
        Test that FLAG=1 produces minimum phase system (zeros in LHP).

        Mathematical property: All zeros should have Re(z) < 0 for continuous-time.
        """
        n_points = 100
        omega = np.linspace(0.1, 10.0, n_points)
        rfrdat = np.zeros(n_points, dtype=float)
        ifrdat = np.zeros(n_points, dtype=float)

        # System with two poles: H(s) = 1/((s+0.5)(s+2))
        for i in range(n_points):
            w = omega[i]
            s = 1j * w
            h = 1.0 / ((s + 0.5) * (s + 2.0))
            rfrdat[i] = np.real(h)
            ifrdat[i] = np.imag(h)

        n = 2
        discfl = 0
        flag = 1  # Enforce minimum phase
        tol = 0.0

        A, B, C, D, N_out, info = sb10yd(discfl, flag, n, rfrdat, ifrdat, omega, tol)

        self.assertEqual(info, 0)

        # All eigenvalues (poles) should be in LHP
        eig_a = np.linalg.eigvals(A)
        self.assertTrue(np.all(np.real(eig_a) < 0), "All poles should be stable")


if __name__ == '__main__':
    unittest.main()



"""
Tests for BB02AD - Benchmark examples for discrete-time algebraic Riccati equations.

BB02AD generates benchmark examples for the numerical solution of discrete-time
algebraic Riccati equations (DAREs) of the form:

    0 = A^T X A - X - (A^T X B + S)(R + B^T X B)^{-1} (B^T X A + S^T) + Q

corresponding to the DAREX collection. This routine is similar to BB01AD but
for discrete-time systems.
"""

import numpy as np
import pytest
from numpy.testing import assert_allclose


def test_bb02ad_example_2_3():
    """
    Test Example 2.3 from SLICOT HTML documentation.

    "increasingly bad scaled system as eps -> oo"

    Uses NR=(2,3), DEF='N', DPAR(1)=0.1234.
    Expected output from HTML doc:
      A = [[0.0, 0.1234], [0.0, 0.0]]
      B = [[0.0], [1.0]]
      Q = I_2
      R = 1.0
      X = [[1.0, 0.0], [0.0, 1.0152]]
    """
    from slicot import bb02ad

    nr = np.array([2, 3], dtype=np.int32)
    dpar = np.array([0.1234, 0.0, 0.0, 0.0], dtype=np.float64)
    ipar = np.array([0, 0, 0], dtype=np.int32)
    bpar = np.array([True, True, True, False, False, True, True], dtype=bool)

    result = bb02ad('N', nr, dpar, ipar, bpar)
    vec, n, m, p, a, b, c, q, r, s, x, info = result

    assert info == 0, f"BB02AD failed with info={info}"
    assert n == 2
    assert m == 1
    assert p == 2

    a_expected = np.array([
        [0.0000, 0.1234],
        [0.0000, 0.0000],
    ], order='F', dtype=np.float64)
    assert_allclose(a[:n, :n], a_expected, rtol=1e-3, atol=1e-4)

    b_expected = np.array([
        [0.0],
        [1.0],
    ], order='F', dtype=np.float64)
    assert_allclose(b[:n, :m], b_expected, rtol=1e-3, atol=1e-4)

    q_expected = np.array([
        [1.0, 0.0],
        [0.0, 1.0],
    ], order='F', dtype=np.float64)
    assert_allclose(q[:n, :n], q_expected, rtol=1e-3, atol=1e-4)

    r_expected = np.array([[1.0]], order='F', dtype=np.float64)
    assert_allclose(r[:m, :m], r_expected, rtol=1e-3, atol=1e-4)

    x_expected = np.array([
        [1.0000, 0.0000],
        [0.0000, 1.0152],
    ], order='F', dtype=np.float64)
    assert_allclose(x[:n, :n], x_expected, rtol=1e-3, atol=1e-4)

    assert vec[0] == True
    assert vec[1] == True
    assert vec[2] == True
    assert vec[3] == True
    assert vec[4] == True
    assert vec[5] == False
    assert vec[8] == True
    assert vec[9] == True


def test_bb02ad_example_1_1_default():
    """
    Test Example 1.1: Van Dooren 1981, Ex. II (singular R matrix).

    Uses default parameters (DEF='D').
    This is a 2x2 problem with known exact solution X = I (identity).

    Note: Example 1.1 has R=0 (singular), so we use BPAR[3]=False to get
    factored form (B, R) instead of trying to compute G = B*R^{-1}*B^T.
    """
    from slicot import bb02ad

    nr = np.array([1, 1], dtype=np.int32)
    dpar = np.zeros(4, dtype=np.float64)
    ipar = np.zeros(3, dtype=np.int32)
    bpar = np.array([True, True, True, False, True, True, False], dtype=bool)

    result = bb02ad('D', nr, dpar, ipar, bpar)
    vec, n, m, p, a, b, c, q, r, s, x, info = result

    assert info == 0, f"BB02AD failed with info={info}"
    assert n == 2
    assert m == 1
    assert p == 1

    x_expected = np.array([
        [1.0, 0.0],
        [0.0, 1.0],
    ], order='F', dtype=np.float64)
    assert_allclose(x[:n, :n], x_expected, rtol=1e-14)

    assert vec[9] == True


def test_bb02ad_example_1_3():
    """
    Test Example 1.3: Jonckheere 1981.

    (A,B) controllable, no solution X <= 0.
    Expected X = [[1, 2], [2, 2+sqrt(5)]]
    """
    from slicot import bb02ad

    nr = np.array([1, 3], dtype=np.int32)
    dpar = np.zeros(4, dtype=np.float64)
    ipar = np.zeros(3, dtype=np.int32)
    bpar = np.array([True, True, True, True, True, True, False], dtype=bool)

    result = bb02ad('D', nr, dpar, ipar, bpar)
    vec, n, m, p, a, b, c, q, r, s, x, info = result

    assert info == 0
    assert n == 2
    assert m == 1
    assert p == 2

    a_expected = np.array([
        [0.0, 1.0],
        [0.0, 0.0],
    ], order='F', dtype=np.float64)
    assert_allclose(a[:n, :n], a_expected, rtol=1e-14)

    assert vec[9] == True
    x_expected = np.array([
        [1.0, 2.0],
        [2.0, 2.0 + np.sqrt(5.0)],
    ], order='F', dtype=np.float64)
    assert_allclose(x[:n, :n], x_expected, rtol=1e-14)


def test_bb02ad_example_2_1():
    """
    Test Example 2.1: Laub 1979, Ex. 2 (uncontrollable-unobservable data).

    Uses DPAR(1) as the scalar R matrix value.
    """
    from slicot import bb02ad

    nr = np.array([2, 1], dtype=np.int32)
    dpar = np.array([1e7, 0.0, 0.0, 0.0], dtype=np.float64)
    ipar = np.zeros(3, dtype=np.int32)
    bpar = np.array([True, True, True, True, True, True, False], dtype=bool)

    result = bb02ad('D', nr, dpar, ipar, bpar)
    vec, n, m, p, a, b, c, q, r, s, x, info = result

    assert info == 0
    assert n == 2
    assert m == 1
    assert p == 2

    assert vec[9] == True


def test_bb02ad_example_4_1():
    """
    Test Example 4.1: Pappas et al. 1980, Ex. 3 (scalable size).

    Uses IPAR(1) to specify problem size.
    """
    from slicot import bb02ad

    nr = np.array([4, 1], dtype=np.int32)
    dpar = np.array([1.0, 0.0, 0.0, 0.0], dtype=np.float64)
    ipar = np.array([5, 0, 0], dtype=np.int32)
    bpar = np.array([True, True, True, True, True, True, False], dtype=bool)

    result = bb02ad('N', nr, dpar, ipar, bpar)
    vec, n, m, p, a, b, c, q, r, s, x, info = result

    assert info == 0
    assert n == 5
    assert m == 1

    assert vec[9] == True

    for i in range(n):
        assert_allclose(x[i, i], float(i + 1), rtol=1e-14)


def test_bb02ad_with_s_matrix():
    """
    Test Example 1.2 which has non-zero S matrix.

    Ionescu/Weiss 1992: singular R matrix, nonzero S matrix.
    Note: R is not singular in this example, but we use BPAR[3]=False
    to get factored form (B, R) and access S.
    """
    from slicot import bb02ad

    nr = np.array([1, 2], dtype=np.int32)
    dpar = np.zeros(4, dtype=np.float64)
    ipar = np.zeros(3, dtype=np.int32)
    bpar = np.array([True, True, True, False, True, True, True], dtype=bool)

    result = bb02ad('D', nr, dpar, ipar, bpar)
    vec, n, m, p, a, b, c, q, r, s, x, info = result

    assert info == 0
    assert n == 2
    assert m == 2

    assert vec[8] == True


def test_bb02ad_factored_q():
    """
    Test requesting factored form for Q (BPAR(1)=False).

    When BPAR(1)=False, the routine returns C and Q0 from Q = C^T Q0 C.
    """
    from slicot import bb02ad

    nr = np.array([1, 3], dtype=np.int32)
    dpar = np.zeros(4, dtype=np.float64)
    ipar = np.zeros(3, dtype=np.int32)
    bpar = np.array([False, True, True, True, True, True, False], dtype=bool)

    result = bb02ad('D', nr, dpar, ipar, bpar)
    vec, n, m, p, a, b, c, q, r, s, x, info = result

    assert info == 0
    assert vec[5] == True


def test_bb02ad_factored_g():
    """
    Test requesting factored form for G (BPAR(4)=False).

    When BPAR(4)=False, the routine returns B and R from G = B R^{-1} B^T.
    """
    from slicot import bb02ad

    nr = np.array([1, 3], dtype=np.int32)
    dpar = np.zeros(4, dtype=np.float64)
    ipar = np.zeros(3, dtype=np.int32)
    bpar = np.array([True, True, True, False, True, True, False], dtype=bool)

    result = bb02ad('D', nr, dpar, ipar, bpar)
    vec, n, m, p, a, b, c, q, r, s, x, info = result

    assert info == 0
    assert vec[4] == True


def test_bb02ad_invalid_nr_group():
    """
    Test error handling for invalid NR(1) parameter.
    """
    from slicot import bb02ad

    nr = np.array([5, 1], dtype=np.int32)
    dpar = np.zeros(4, dtype=np.float64)
    ipar = np.zeros(3, dtype=np.int32)
    bpar = np.array([True, True, True, True, True, True, False], dtype=bool)

    result = bb02ad('D', nr, dpar, ipar, bpar)
    vec, n, m, p, a, b, c, q, r, s, x, info = result

    assert info == -2


def test_bb02ad_invalid_example_number():
    """
    Test error handling for invalid example number in group.
    """
    from slicot import bb02ad

    nr = np.array([1, 20], dtype=np.int32)
    dpar = np.zeros(4, dtype=np.float64)
    ipar = np.zeros(3, dtype=np.int32)
    bpar = np.array([True, True, True, True, True, True, False], dtype=bool)

    result = bb02ad('D', nr, dpar, ipar, bpar)
    vec, n, m, p, a, b, c, q, r, s, x, info = result

    assert info == -2


def test_bb02ad_division_by_zero():
    """
    Test error handling when division by zero would occur.

    Example 2.2 with DPAR(1)=0 causes division by zero.
    """
    from slicot import bb02ad

    nr = np.array([2, 2], dtype=np.int32)
    dpar = np.array([0.0, 0.0, 0.0, 0.0], dtype=np.float64)
    ipar = np.zeros(3, dtype=np.int32)
    bpar = np.array([True, True, True, True, True, True, False], dtype=bool)

    result = bb02ad('N', nr, dpar, ipar, bpar)
    vec, n, m, p, a, b, c, q, r, s, x, info = result

    assert info == 2


def test_bb02ad_dare_residual():
    """
    Validate mathematical property: DARE residual for example with known X.

    For the DARE (when S=0):
        0 = A^T X A - X - A^T X B (R + B^T X B)^{-1} B^T X A + Q

    The solution X should satisfy this equation.
    """
    from slicot import bb02ad

    nr = np.array([1, 3], dtype=np.int32)
    dpar = np.zeros(4, dtype=np.float64)
    ipar = np.zeros(3, dtype=np.int32)
    bpar = np.array([True, True, True, False, True, True, False], dtype=bool)

    result = bb02ad('D', nr, dpar, ipar, bpar)
    vec, n, m, p, a, b, c, q, r, s, x, info = result

    assert info == 0
    assert vec[9] == True

    a_out = a[:n, :n]
    b_out = b[:n, :m]
    q_out = q[:n, :n]
    r_out = r[:m, :m]
    x_out = x[:n, :n]

    atxa = a_out.T @ x_out @ a_out
    btxb = b_out.T @ x_out @ b_out
    r_plus_btxb = r_out + btxb
    atxb = a_out.T @ x_out @ b_out
    btxa = b_out.T @ x_out @ a_out

    feedback_term = atxb @ np.linalg.solve(r_plus_btxb, btxa)
    residual = atxa - x_out - feedback_term + q_out

    assert_allclose(residual, np.zeros((n, n)), atol=1e-10)


def test_bb02ad_solution_symmetry():
    """
    Validate mathematical property: solution X is symmetric.

    The solution to a DARE should be symmetric: X = X^T.
    """
    from slicot import bb02ad

    nr = np.array([2, 1], dtype=np.int32)
    dpar = np.array([1e6, 0.0, 0.0, 0.0], dtype=np.float64)
    ipar = np.zeros(3, dtype=np.int32)
    bpar = np.array([True, True, True, True, True, True, False], dtype=bool)

    result = bb02ad('D', nr, dpar, ipar, bpar)
    vec, n, m, p, a, b, c, q, r, s, x, info = result

    assert info == 0
    assert vec[9] == True

    x_out = x[:n, :n]
    assert_allclose(x_out, x_out.T, rtol=1e-14)


def test_bb02ad_q_symmetry():
    """
    Validate mathematical property: Q matrix is symmetric.

    Q is always symmetric in the DARE formulation.
    """
    from slicot import bb02ad

    nr = np.array([1, 3], dtype=np.int32)
    dpar = np.zeros(4, dtype=np.float64)
    ipar = np.zeros(3, dtype=np.int32)
    bpar = np.array([True, True, True, True, True, True, False], dtype=bool)

    result = bb02ad('D', nr, dpar, ipar, bpar)
    vec, n, m, p, a, b, c, q, r, s, x, info = result

    assert info == 0

    q_out = q[:n, :n]
    assert_allclose(q_out, q_out.T, rtol=1e-14)


def test_bb02ad_r_symmetry():
    """
    Validate mathematical property: R matrix is symmetric.

    R is always symmetric in the DARE formulation.
    """
    from slicot import bb02ad

    nr = np.array([1, 2], dtype=np.int32)
    dpar = np.zeros(4, dtype=np.float64)
    ipar = np.zeros(3, dtype=np.int32)
    bpar = np.array([True, True, True, False, True, True, False], dtype=bool)

    result = bb02ad('D', nr, dpar, ipar, bpar)
    vec, n, m, p, a, b, c, q, r, s, x, info = result

    assert info == 0

    r_out = r[:m, :m]
    assert_allclose(r_out, r_out.T, rtol=1e-14)



"""
Tests for UD01MD: Printing a real matrix.

UD01MD prints an M-by-N real matrix A row by row to a string with
7 significant figures. In Python, returns a formatted string instead
of writing to a file.

Format:
- Title line: "Matrix A (MxN)"
- Column headers every L columns
- Values in scientific notation with 7 significant figures
"""

import numpy as np
import pytest


def test_ud01md_basic():
    """
    Validate basic functionality using SLICOT HTML doc example.

    From HTML doc:
    - M=4, N=4, L=4 (4x4 matrix, 4 elements per line)
    - Matrix values 1-16
    """
    from slicot import ud01md

    m, n, l = 4, 4, 4

    a = np.array([
        [1.0, 2.0, 3.0, 4.0],
        [5.0, 6.0, 7.0, 8.0],
        [9.0, 10.0, 11.0, 12.0],
        [13.0, 14.0, 15.0, 16.0]
    ], order='F', dtype=float)

    text = 'Matrix A'

    result, info = ud01md(m, n, l, a, text)

    assert info == 0
    assert isinstance(result, str)

    assert 'Matrix A' in result
    assert '4' in result

    lines = result.strip().split('\n')
    assert len(lines) >= m + 2


def test_ud01md_small_l():
    """
    Test with L=2 (2 elements per line, requiring column blocks).

    This tests the column blocking feature where N > L.
    """
    from slicot import ud01md

    m, n, l = 2, 4, 2

    a = np.array([
        [1.0, 2.0, 3.0, 4.0],
        [5.0, 6.0, 7.0, 8.0]
    ], order='F', dtype=float)

    text = 'Test Matrix'

    result, info = ud01md(m, n, l, a, text)

    assert info == 0
    assert 'Test Matrix' in result


def test_ud01md_single_column():
    """
    Test edge case: single column matrix (N=1).
    """
    from slicot import ud01md

    m, n, l = 3, 1, 1

    a = np.array([
        [1.0],
        [2.0],
        [3.0]
    ], order='F', dtype=float)

    text = 'Column Vector'

    result, info = ud01md(m, n, l, a, text)

    assert info == 0
    assert 'Column Vector' in result
    assert '3' in result
    assert '1' in result


def test_ud01md_single_row():
    """
    Test edge case: single row matrix (M=1).
    """
    from slicot import ud01md

    m, n, l = 1, 4, 4

    a = np.array([
        [1.0, 2.0, 3.0, 4.0]
    ], order='F', dtype=float)

    text = 'Row Vector'

    result, info = ud01md(m, n, l, a, text)

    assert info == 0
    assert 'Row Vector' in result


def test_ud01md_l_equals_5():
    """
    Test with maximum allowed L=5.
    """
    from slicot import ud01md

    m, n, l = 2, 5, 5

    a = np.array([
        [1.0, 2.0, 3.0, 4.0, 5.0],
        [6.0, 7.0, 8.0, 9.0, 10.0]
    ], order='F', dtype=float)

    text = 'Max L Matrix'

    result, info = ud01md(m, n, l, a, text)

    assert info == 0
    assert 'Max L Matrix' in result


def test_ud01md_scientific_notation():
    """
    Test that values are formatted in scientific notation with 7 figures.

    Random seed: 42 (for reproducibility)
    """
    from slicot import ud01md

    np.random.seed(42)

    m, n, l = 2, 2, 2

    a = np.array([
        [1.234567e5, 9.876543e-3],
        [5.555555e10, 1.111111e-8]
    ], order='F', dtype=float)

    text = 'Scientific'

    result, info = ud01md(m, n, l, a, text)

    assert info == 0
    assert 'D' in result or 'E' in result or 'e' in result


def test_ud01md_error_m_invalid():
    """
    Test error: M < 1.
    """
    from slicot import ud01md

    m, n, l = 0, 4, 4

    a = np.array([[1.0, 2.0, 3.0, 4.0]], order='F', dtype=float)

    text = 'Test'

    result, info = ud01md(m, n, l, a, text)

    assert info == -1


def test_ud01md_error_n_invalid():
    """
    Test error: N < 1.
    """
    from slicot import ud01md

    m, n, l = 4, 0, 4

    a = np.array([[1.0], [2.0], [3.0], [4.0]], order='F', dtype=float)

    text = 'Test'

    result, info = ud01md(m, n, l, a, text)

    assert info == -2


def test_ud01md_error_l_too_small():
    """
    Test error: L < 1.
    """
    from slicot import ud01md

    m, n, l = 4, 4, 0

    a = np.array([
        [1.0, 2.0, 3.0, 4.0],
        [5.0, 6.0, 7.0, 8.0],
        [9.0, 10.0, 11.0, 12.0],
        [13.0, 14.0, 15.0, 16.0]
    ], order='F', dtype=float)

    text = 'Test'

    result, info = ud01md(m, n, l, a, text)

    assert info == -3


def test_ud01md_error_l_too_large():
    """
    Test error: L > 5.
    """
    from slicot import ud01md

    m, n, l = 4, 4, 6

    a = np.array([
        [1.0, 2.0, 3.0, 4.0],
        [5.0, 6.0, 7.0, 8.0],
        [9.0, 10.0, 11.0, 12.0],
        [13.0, 14.0, 15.0, 16.0]
    ], order='F', dtype=float)

    text = 'Test'

    result, info = ud01md(m, n, l, a, text)

    assert info == -3


def test_ud01md_long_text():
    """
    Test with long text title (truncated to 72 chars).
    """
    from slicot import ud01md

    m, n, l = 2, 2, 2

    a = np.array([
        [1.0, 2.0],
        [3.0, 4.0]
    ], order='F', dtype=float)

    text = 'A' * 100

    result, info = ud01md(m, n, l, a, text)

    assert info == 0


def test_ud01md_empty_text():
    """
    Test with empty text title.
    """
    from slicot import ud01md

    m, n, l = 2, 2, 2

    a = np.array([
        [1.0, 2.0],
        [3.0, 4.0]
    ], order='F', dtype=float)

    text = ''

    result, info = ud01md(m, n, l, a, text)

    assert info == 0



"""
Tests for SG03BZ - Full generalized Lyapunov solver (complex).

Computes Cholesky factor U of X = op(U)^H * op(U) for:

Continuous-time (DICO='C'):
    op(A)^H * X * op(E) + op(E)^H * X * op(A) = -SCALE^2 * op(B)^H * op(B)

Discrete-time (DICO='D'):
    op(A)^H * X * op(A) - op(E)^H * X * op(E) = -SCALE^2 * op(B)^H * op(B)

where op(K) = K (TRANS='N') or op(K) = K^H (TRANS='C').

Unlike SG03BS/SG03BT which work on triangular Schur form, SG03BZ:
1. Performs QZ factorization if FACT='N'
2. Transforms B appropriately
3. Calls SG03BS (discrete) or SG03BT (continuous) on reduced problem
4. Transforms solution back

Key test cases:
- FACT='N' (compute Schur factorization)
- FACT='F' (factorization supplied)
- DICO='C' (continuous-time, c-stable: eigenvalues with Re < 0)
- DICO='D' (discrete-time, d-stable: eigenvalue modulus < 1)
- TRANS='N' and TRANS='C'
- Identity Q/Z detection
- Various M vs N relationships (M < N, M >= N)
"""

import numpy as np
from slicot import sg03bz


def test_sg03bz_continuous_trans_n_fact_n():
    """
    Test SG03BZ with DICO='C', TRANS='N', FACT='N'.

    Continuous-time equation with no pre-factorization:
        A^H * X * E + E^H * X * A = -SCALE^2 * B^H * B
    where X = U^H * U.

    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    n = 3

    a = np.array([
        [-0.5+0.1j, 0.2+0.05j, 0.1+0.02j],
        [0.1-0.05j, -0.4-0.1j, 0.15+0.03j],
        [0.05+0.02j, 0.08-0.04j, -0.3+0.05j]
    ], dtype=complex, order='F')

    e = np.array([
        [1.0+0j, 0.1+0.02j, 0.05+0.01j],
        [0.0, 1.0+0j, 0.08+0.02j],
        [0.0, 0.0, 1.0+0j]
    ], dtype=complex, order='F')

    b = np.array([
        [0.5, 0.2+0.1j, 0.1+0.05j],
        [0.0, 0.3, 0.15+0.08j]
    ], dtype=complex, order='F')

    a_orig = a.copy()
    e_orig = e.copy()
    b_orig = b.copy()

    q = np.zeros((n, n), dtype=complex, order='F')
    z = np.zeros((n, n), dtype=complex, order='F')

    u, scale, _alpha, _beta, info = sg03bz('C', 'N', 'N', a, e, q, z, b)

    assert info == 0, f"SG03BZ failed with INFO={info}"
    assert 0.0 < scale <= 1.0, f"Invalid scale: {scale}"

    np.testing.assert_allclose(np.tril(u, -1), 0, atol=1e-14)

    x = u.conj().T @ u
    lhs = a_orig.conj().T @ x @ e_orig + e_orig.conj().T @ x @ a_orig
    rhs = -scale**2 * (b_orig.conj().T @ b_orig)
    residual = np.linalg.norm(lhs - rhs) / max(np.linalg.norm(lhs), np.linalg.norm(rhs), 1e-15)
    assert residual < 1e-9, f"Lyapunov residual too large: {residual}"

    assert np.all(np.diag(u).real >= -1e-14), "U diagonal should be non-negative real"


def test_sg03bz_continuous_trans_c_fact_n():
    """
    Test SG03BZ with DICO='C', TRANS='C', FACT='N'.

    Continuous-time conjugate transpose equation:
        A * X * E^H + E * X * A^H = -SCALE^2 * B * B^H
    where X = U * U^H.

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)
    n = 3
    m = 2

    a = np.array([
        [-0.4+0.08j, 0.15+0.04j, 0.08+0.02j],
        [0.1-0.04j, -0.35-0.08j, 0.12+0.03j],
        [0.04+0.02j, 0.06-0.03j, -0.25+0.04j]
    ], dtype=complex, order='F')

    e = np.array([
        [1.0+0j, 0.12+0.03j, 0.06+0.02j],
        [0.0, 1.0+0j, 0.1+0.02j],
        [0.0, 0.0, 1.0+0j]
    ], dtype=complex, order='F')

    b = np.array([
        [0.4+0j, 0.0+0j, 0.0+0j],
        [0.18+0.08j, 0.35+0j, 0.0+0j],
        [0.09+0.04j, 0.12+0.06j, 0.2+0j]
    ], dtype=complex, order='F')

    a_orig = a.copy()
    e_orig = e.copy()
    b_orig = b.copy()

    q = np.zeros((n, n), dtype=complex, order='F')
    z = np.zeros((n, n), dtype=complex, order='F')

    u, scale, _alpha, _beta, info = sg03bz('C', 'N', 'C', a, e, q, z, b)

    assert info == 0, f"SG03BZ failed with INFO={info}"
    assert 0.0 < scale <= 1.0, f"Invalid scale: {scale}"

    np.testing.assert_allclose(np.tril(u, -1), 0, atol=1e-14)

    x = u @ u.conj().T
    lhs = a_orig @ x @ e_orig.conj().T + e_orig @ x @ a_orig.conj().T
    rhs = -scale**2 * (b_orig @ b_orig.conj().T)
    residual = np.linalg.norm(lhs - rhs) / max(np.linalg.norm(lhs), np.linalg.norm(rhs), 1e-15)
    assert residual < 1e-9, f"Lyapunov residual too large: {residual}"


def test_sg03bz_discrete_trans_n_fact_n():
    """
    Test SG03BZ with DICO='D', TRANS='N', FACT='N'.

    Discrete-time equation:
        A^H * X * A - E^H * X * E = -SCALE^2 * B^H * B
    where X = U^H * U.

    Random seed: 456 (for reproducibility)
    """
    np.random.seed(456)
    n = 3
    m = 2

    a = np.array([
        [0.3+0.1j, 0.1+0.05j, 0.05+0.02j],
        [0.05-0.02j, 0.2-0.08j, 0.08+0.03j],
        [0.02+0.01j, 0.04-0.02j, 0.15+0.05j]
    ], dtype=complex, order='F')

    e = np.array([
        [1.0+0j, 0.1+0.02j, 0.05+0.01j],
        [0.0, 1.0+0j, 0.08+0.02j],
        [0.0, 0.0, 1.0+0j]
    ], dtype=complex, order='F')

    b = np.array([
        [0.5, 0.2+0.1j, 0.1+0.05j],
        [0.0, 0.3, 0.15+0.08j]
    ], dtype=complex, order='F')

    a_orig = a.copy()
    e_orig = e.copy()
    b_orig = b.copy()

    q = np.zeros((n, n), dtype=complex, order='F')
    z = np.zeros((n, n), dtype=complex, order='F')

    u, scale, _alpha, _beta, info = sg03bz('D', 'N', 'N', a, e, q, z, b)

    assert info == 0, f"SG03BZ failed with INFO={info}"
    assert 0.0 < scale <= 1.0, f"Invalid scale: {scale}"

    np.testing.assert_allclose(np.tril(u, -1), 0, atol=1e-14)

    x = u.conj().T @ u
    lhs = a_orig.conj().T @ x @ a_orig - e_orig.conj().T @ x @ e_orig
    rhs = -scale**2 * (b_orig.conj().T @ b_orig)
    residual = np.linalg.norm(lhs - rhs) / max(np.linalg.norm(lhs), np.linalg.norm(rhs), 1e-15)
    assert residual < 1e-9, f"Lyapunov residual too large: {residual}"


def test_sg03bz_discrete_trans_c_fact_n():
    """
    Test SG03BZ with DICO='D', TRANS='C', FACT='N'.

    Discrete-time conjugate transpose equation:
        A * X * A^H - E * X * E^H = -SCALE^2 * B * B^H
    where X = U * U^H.

    Random seed: 789 (for reproducibility)
    """
    np.random.seed(789)
    n = 3
    m = 2

    a = np.array([
        [0.25+0.08j, 0.12+0.04j, 0.06+0.02j],
        [0.04-0.02j, 0.18-0.06j, 0.09+0.03j],
        [0.02+0.01j, 0.03-0.02j, 0.12+0.04j]
    ], dtype=complex, order='F')

    e = np.array([
        [1.0+0j, 0.15+0.03j, 0.08+0.02j],
        [0.0, 1.0+0j, 0.1+0.02j],
        [0.0, 0.0, 1.0+0j]
    ], dtype=complex, order='F')

    b = np.array([
        [0.4+0j, 0.0+0j, 0.0+0j],
        [0.18+0.08j, 0.35+0j, 0.0+0j],
        [0.09+0.04j, 0.12+0.06j, 0.2+0j]
    ], dtype=complex, order='F')

    a_orig = a.copy()
    e_orig = e.copy()
    b_orig = b.copy()

    q = np.zeros((n, n), dtype=complex, order='F')
    z = np.zeros((n, n), dtype=complex, order='F')

    u, scale, _alpha, _beta, info = sg03bz('D', 'N', 'C', a, e, q, z, b)

    assert info == 0, f"SG03BZ failed with INFO={info}"
    assert 0.0 < scale <= 1.0, f"Invalid scale: {scale}"

    np.testing.assert_allclose(np.tril(u, -1), 0, atol=1e-14)

    x = u @ u.conj().T
    lhs = a_orig @ x @ a_orig.conj().T - e_orig @ x @ e_orig.conj().T
    rhs = -scale**2 * (b_orig @ b_orig.conj().T)
    residual = np.linalg.norm(lhs - rhs) / max(np.linalg.norm(lhs), np.linalg.norm(rhs), 1e-15)
    assert residual < 1e-9, f"Lyapunov residual too large: {residual}"


def test_sg03bz_continuous_fact_f():
    """
    Test SG03BZ with FACT='F' (factorization supplied).

    When FACT='F', A and E must already be in upper triangular (Schur) form,
    and Q, Z must be the unitary transformation matrices.

    Random seed: 111 (for reproducibility)
    """
    np.random.seed(111)
    n = 3
    m = 2

    a = np.array([
        [-0.5+0.1j, 0.1+0.05j, 0.05+0.02j],
        [0.0, -0.3-0.1j, 0.08+0.03j],
        [0.0, 0.0, -0.2+0.05j]
    ], dtype=complex, order='F')

    e = np.array([
        [1.0+0j, 0.1+0.02j, 0.05+0.01j],
        [0.0, 1.0+0j, 0.08+0.02j],
        [0.0, 0.0, 1.0+0j]
    ], dtype=complex, order='F')

    q = np.eye(n, dtype=complex, order='F')
    z = np.eye(n, dtype=complex, order='F')

    b = np.array([
        [0.5, 0.2+0.1j, 0.1+0.05j],
        [0.0, 0.3, 0.15+0.08j]
    ], dtype=complex, order='F')

    a_orig = a.copy()
    e_orig = e.copy()
    b_orig = b.copy()

    u, scale, _alpha, _beta, info = sg03bz('C', 'F', 'N', a, e, q, z, b)

    assert info == 0, f"SG03BZ failed with INFO={info}"
    assert 0.0 < scale <= 1.0, f"Invalid scale: {scale}"

    x = u.conj().T @ u
    lhs = a_orig.conj().T @ x @ e_orig + e_orig.conj().T @ x @ a_orig
    rhs = -scale**2 * (b_orig.conj().T @ b_orig)
    residual = np.linalg.norm(lhs - rhs) / max(np.linalg.norm(lhs), np.linalg.norm(rhs), 1e-15)
    assert residual < 1e-9, f"Lyapunov residual too large: {residual}"


def test_sg03bz_m_equals_n():
    """
    Test SG03BZ when M equals N.

    Random seed: 222 (for reproducibility)
    """
    np.random.seed(222)
    n = 3
    m = 3

    a = np.array([
        [-0.4+0.1j, 0.15+0.04j, 0.08+0.02j],
        [0.1-0.03j, -0.35-0.08j, 0.12+0.03j],
        [0.05+0.02j, 0.06-0.03j, -0.25+0.04j]
    ], dtype=complex, order='F')

    e = np.eye(n, dtype=complex, order='F')

    b = np.array([
        [0.5, 0.2+0.1j, 0.1+0.05j],
        [0.0, 0.4, 0.15+0.08j],
        [0.0, 0.0, 0.3]
    ], dtype=complex, order='F')

    a_orig = a.copy()
    e_orig = e.copy()
    b_orig = b.copy()

    q = np.zeros((n, n), dtype=complex, order='F')
    z = np.zeros((n, n), dtype=complex, order='F')

    u, scale, _alpha, _beta, info = sg03bz('C', 'N', 'N', a, e, q, z, b)

    assert info == 0, f"SG03BZ failed with INFO={info}"

    x = u.conj().T @ u
    lhs = a_orig.conj().T @ x @ e_orig + e_orig.conj().T @ x @ a_orig
    rhs = -scale**2 * (b_orig.conj().T @ b_orig)
    residual = np.linalg.norm(lhs - rhs) / max(np.linalg.norm(lhs), np.linalg.norm(rhs), 1e-15)
    assert residual < 1e-9, f"Lyapunov residual too large: {residual}"


def test_sg03bz_m_greater_n():
    """
    Test SG03BZ when M > N.

    Random seed: 333 (for reproducibility)
    """
    np.random.seed(333)
    n = 3
    m = 5

    a = np.array([
        [-0.4+0.1j, 0.15+0.04j, 0.08+0.02j],
        [0.1-0.03j, -0.35-0.08j, 0.12+0.03j],
        [0.05+0.02j, 0.06-0.03j, -0.25+0.04j]
    ], dtype=complex, order='F')

    e = np.eye(n, dtype=complex, order='F')

    b = np.zeros((m, n), dtype=complex, order='F')
    for i in range(m):
        for j in range(n):
            b[i, j] = 0.1 * (np.random.randn() + 1j * np.random.randn())
    b = np.asfortranarray(b)

    a_orig = a.copy()
    e_orig = e.copy()
    b_orig = b.copy()

    q = np.zeros((n, n), dtype=complex, order='F')
    z = np.zeros((n, n), dtype=complex, order='F')

    u, scale, _alpha, _beta, info = sg03bz('C', 'N', 'N', a, e, q, z, b)

    assert info == 0, f"SG03BZ failed with INFO={info}"

    x = u.conj().T @ u
    lhs = a_orig.conj().T @ x @ e_orig + e_orig.conj().T @ x @ a_orig
    rhs = -scale**2 * (b_orig.conj().T @ b_orig)
    residual = np.linalg.norm(lhs - rhs) / max(np.linalg.norm(lhs), np.linalg.norm(rhs), 1e-15)
    assert residual < 1e-9, f"Lyapunov residual too large: {residual}"


def test_sg03bz_n0():
    """Test SG03BZ with N=0 (quick return)."""
    n = 0
    m = 0

    a = np.zeros((0, 0), dtype=complex, order='F')
    e = np.zeros((0, 0), dtype=complex, order='F')
    q = np.zeros((0, 0), dtype=complex, order='F')
    z = np.zeros((0, 0), dtype=complex, order='F')
    b = np.zeros((0, 0), dtype=complex, order='F')

    u, scale, _alpha, _beta, info = sg03bz('C', 'N', 'N', a, e, q, z, b)

    assert info == 0
    assert scale == 1.0


def test_sg03bz_m0():
    """Test SG03BZ with M=0 (U should be zero)."""
    n = 3
    m = 0

    a = np.array([
        [-0.5+0.1j, 0.1+0.05j, 0.05+0.02j],
        [0.0, -0.3-0.1j, 0.08+0.03j],
        [0.0, 0.0, -0.2+0.05j]
    ], dtype=complex, order='F')

    e = np.eye(n, dtype=complex, order='F')

    q = np.zeros((n, n), dtype=complex, order='F')
    z = np.zeros((n, n), dtype=complex, order='F')
    b = np.zeros((0, n), dtype=complex, order='F')

    u, scale, _alpha, _beta, info = sg03bz('C', 'N', 'N', a, e, q, z, b)

    assert info == 0
    np.testing.assert_allclose(u, 0, atol=1e-14)


def test_sg03bz_continuous_unstable():
    """Test SG03BZ returns INFO=5 for c-unstable pencil (DICO='C')."""
    n = 2
    m = 2

    a = np.array([
        [0.5+0j, 0.1+0.05j],
        [0.0, 0.3-0.1j]
    ], dtype=complex, order='F')

    e = np.eye(n, dtype=complex, order='F')

    q = np.zeros((n, n), dtype=complex, order='F')
    z = np.zeros((n, n), dtype=complex, order='F')

    b = np.array([
        [0.5, 0.2+0.1j],
        [0.0, 0.3]
    ], dtype=complex, order='F')

    u, scale, _alpha, _beta, info = sg03bz('C', 'N', 'N', a, e, q, z, b)

    assert info == 5, f"Expected INFO=5 for c-unstable system, got {info}"


def test_sg03bz_discrete_unstable():
    """Test SG03BZ returns INFO=6 for d-unstable pencil (DICO='D')."""
    n = 2
    m = 2

    a = np.array([
        [1.5+0j, 0.1+0.05j],
        [0.0, 1.2-0.1j]
    ], dtype=complex, order='F')

    e = np.eye(n, dtype=complex, order='F')

    q = np.zeros((n, n), dtype=complex, order='F')
    z = np.zeros((n, n), dtype=complex, order='F')

    b = np.array([
        [0.5, 0.2+0.1j],
        [0.0, 0.3]
    ], dtype=complex, order='F')

    u, scale, _alpha, _beta, info = sg03bz('D', 'N', 'N', a, e, q, z, b)

    assert info == 6, f"Expected INFO=6 for d-unstable system, got {info}"


def test_sg03bz_invalid_dico():
    """Test SG03BZ returns INFO=-1 for invalid DICO."""
    n = 2
    m = 2

    a = np.array([
        [-0.5+0.1j, 0.1+0.05j],
        [0.0, -0.3-0.1j]
    ], dtype=complex, order='F')

    e = np.eye(n, dtype=complex, order='F')
    q = np.zeros((n, n), dtype=complex, order='F')
    z = np.zeros((n, n), dtype=complex, order='F')

    b = np.array([
        [0.5, 0.2+0.1j],
        [0.0, 0.3]
    ], dtype=complex, order='F')

    u, scale, _alpha, _beta, info = sg03bz('X', 'N', 'N', a, e, q, z, b)

    assert info == -1, f"Expected INFO=-1 for invalid DICO, got {info}"


def test_sg03bz_invalid_fact():
    """Test SG03BZ returns INFO=-2 for invalid FACT."""
    n = 2
    m = 2

    a = np.array([
        [-0.5+0.1j, 0.1+0.05j],
        [0.0, -0.3-0.1j]
    ], dtype=complex, order='F')

    e = np.eye(n, dtype=complex, order='F')
    q = np.zeros((n, n), dtype=complex, order='F')
    z = np.zeros((n, n), dtype=complex, order='F')

    b = np.array([
        [0.5, 0.2+0.1j],
        [0.0, 0.3]
    ], dtype=complex, order='F')

    u, scale, _alpha, _beta, info = sg03bz('C', 'X', 'N', a, e, q, z, b)

    assert info == -2, f"Expected INFO=-2 for invalid FACT, got {info}"


def test_sg03bz_invalid_trans():
    """Test SG03BZ returns INFO=-3 for invalid TRANS."""
    n = 2
    m = 2

    a = np.array([
        [-0.5+0.1j, 0.1+0.05j],
        [0.0, -0.3-0.1j]
    ], dtype=complex, order='F')

    e = np.eye(n, dtype=complex, order='F')
    q = np.zeros((n, n), dtype=complex, order='F')
    z = np.zeros((n, n), dtype=complex, order='F')

    b = np.array([
        [0.5, 0.2+0.1j],
        [0.0, 0.3]
    ], dtype=complex, order='F')

    u, scale, _alpha, _beta, info = sg03bz('C', 'N', 'X', a, e, q, z, b)

    assert info == -3, f"Expected INFO=-3 for invalid TRANS, got {info}"


def test_sg03bz_eigenvalue_output():
    """
    Test that alpha/beta eigenvalue outputs are correct.

    For triangular A, E (FACT='F'), eigenvalues should match diagonal entries.

    Random seed: 555 (for reproducibility)
    """
    np.random.seed(555)
    n = 3
    m = 2

    a = np.array([
        [-0.5+0.1j, 0.1+0.05j, 0.05+0.02j],
        [0.0, -0.3-0.1j, 0.08+0.03j],
        [0.0, 0.0, -0.2+0.05j]
    ], dtype=complex, order='F')

    e = np.array([
        [1.0+0j, 0.1+0.02j, 0.05+0.01j],
        [0.0, 0.8+0j, 0.08+0.02j],
        [0.0, 0.0, 0.6+0j]
    ], dtype=complex, order='F')

    q = np.eye(n, dtype=complex, order='F')
    z = np.eye(n, dtype=complex, order='F')

    b = np.array([
        [0.5, 0.2+0.1j, 0.1+0.05j],
        [0.0, 0.3, 0.15+0.08j]
    ], dtype=complex, order='F')

    u, scale, alpha, beta, info = sg03bz('C', 'F', 'N', a, e, q, z, b)

    assert info == 0, f"SG03BZ failed with INFO={info}"

    np.testing.assert_allclose(alpha, np.diag(a), rtol=1e-14)
    np.testing.assert_allclose(beta, np.diag(e), rtol=1e-14)


def test_sg03bz_large_system():
    """
    Test SG03BZ with larger system.

    Random seed: 2024 (for reproducibility)
    """
    np.random.seed(2024)
    n = 8
    m = 4

    diag_a = -0.3 - 0.2 * np.random.rand(n) + 0.1j * np.random.randn(n)
    a = np.diag(diag_a).astype(complex, order='F')
    for i in range(n):
        for j in range(i+1, n):
            a[i, j] = 0.05 * (np.random.randn() + 1j * np.random.randn())
    for i in range(1, n):
        for j in range(i):
            a[i, j] = 0.02 * (np.random.randn() + 1j * np.random.randn())
    a = np.asfortranarray(a)

    e = np.eye(n, dtype=complex, order='F')
    for i in range(n):
        for j in range(i+1, n):
            e[i, j] = 0.02 * (np.random.randn() + 1j * np.random.randn())
    e = np.asfortranarray(e)

    b = 0.1 * (np.random.randn(m, n) + 1j * np.random.randn(m, n))
    b = np.asfortranarray(b.astype(complex))

    a_orig = a.copy()
    e_orig = e.copy()
    b_orig = b.copy()

    q = np.zeros((n, n), dtype=complex, order='F')
    z = np.zeros((n, n), dtype=complex, order='F')

    u, scale, _alpha, _beta, info = sg03bz('C', 'N', 'N', a, e, q, z, b)

    assert info == 0, f"SG03BZ failed with INFO={info}"
    assert 0.0 < scale <= 1.0

    x = u.conj().T @ u
    lhs = a_orig.conj().T @ x @ e_orig + e_orig.conj().T @ x @ a_orig
    rhs = -scale**2 * (b_orig.conj().T @ b_orig)
    residual = np.linalg.norm(lhs - rhs, 'fro') / max(np.linalg.norm(lhs, 'fro'), 1e-15)
    assert residual < 1e-8, f"Lyapunov residual too large: {residual}"


def test_sg03bz_positive_semidefinite_solution():
    """
    Validate X = U^H * U is positive semidefinite.

    Random seed: 666 (for reproducibility)
    """
    np.random.seed(666)
    n = 4
    m = 3

    diag_a = -0.3 - 0.2 * np.random.rand(n) + 0.1j * np.random.randn(n)
    a = np.diag(diag_a).astype(complex, order='F')
    for i in range(n):
        for j in range(i+1, n):
            a[i, j] = 0.05 * (np.random.randn() + 1j * np.random.randn())
    a = np.asfortranarray(a)

    e = np.eye(n, dtype=complex, order='F')

    b = 0.2 * (np.random.randn(m, n) + 1j * np.random.randn(m, n))
    b = np.asfortranarray(b.astype(complex))

    q = np.zeros((n, n), dtype=complex, order='F')
    z = np.zeros((n, n), dtype=complex, order='F')

    u, scale, _alpha, _beta, info = sg03bz('C', 'N', 'N', a.copy(), e.copy(), q, z, b.copy())

    assert info == 0, f"SG03BZ failed with INFO={info}"

    np.testing.assert_allclose(np.tril(u, -1), 0, atol=1e-14)

    x = u.conj().T @ u

    eig_x = np.linalg.eigvalsh(x)
    assert np.all(eig_x >= -1e-10), f"X should be positive semidefinite, min eig={eig_x.min()}"



"""
Tests for MB03XZ: Eigenvalues of complex Hamiltonian matrix.

MB03XZ computes the eigenvalues of a complex Hamiltonian matrix:
    H = [ A   G  ]    G = G^H, Q = Q^H
        [ Q  -A^H]

Uses embedding to real skew-Hamiltonian and structured Schur form.
"""

import sys

import numpy as np
import pytest
from numpy.testing import assert_allclose


def test_mb03xz_basic_eigenvalues():
    """
    Test basic eigenvalue computation using HTML doc example.

    Input data from SLICOT MB03XZ.html example:
    N=4, BALANC='N', JOB='G', JOBU='U'
    """
    from slicot import mb03xz

    n = 4
    # A matrix from HTML doc (row-wise as displayed)
    a = np.array([
        [0.8147+0.4217j, 0.6323+0.6557j, 0.9575+0.6787j, 0.9571+0.6554j],
        [0.9057+0.9157j, 0.0975+0.0357j, 0.9648+0.7577j, 0.4853+0.1711j],
        [0.1269+0.7922j, 0.2784+0.8491j, 0.1576+0.7431j, 0.8002+0.7060j],
        [0.9133+0.9594j, 0.5468+0.9339j, 0.9705+0.3922j, 0.1418+0.0318j],
    ], dtype=complex, order='F')

    # QG matrix (lower triangular Q in cols 1:N, upper triangular G in cols 2:N+1)
    # Parsed from HTML doc data
    qg = np.array([
        [0.2769+0.0j,      0.6948+0.0j,      0.4387+0.7513j, 0.1869+0.8909j, 0.7094+0.1493j],
        [0.0462+0.1626j,   0.3171+0.0j,      0.3816+0.0j,    0.4898+0.9593j, 0.7547+0.2575j],
        [0.0971+0.1190j,   0.9502+0.5853j,   0.7655+0.0j,    0.4456+0.0j,    0.2760+0.8407j],
        [0.8235+0.4984j,   0.0344+0.2238j,   0.7952+0.6991j, 0.6463+0.0j,    0.6797+0.0j],
    ], dtype=complex, order='F')

    # Expected eigenvalues from HTML doc
    expected_wr = np.array([3.0844, -3.0844, 0.5241, -0.5241, 0.8824, -0.8824, 0.4459, -0.4459])
    expected_wi = np.array([2.7519,  2.7519, -1.3026, -1.3026, -0.6918, -0.6918, 0.4748, 0.4748])

    wr, wi, sc, gc, u1, u2, ilo, scale, info = mb03xz(
        a, qg, balanc='N', job='G', jobu='U'
    )

    assert info == 0
    # Eigenvalues should match (may be in different order, so sort by magnitude)
    eig_computed = wr + 1j * wi
    eig_expected = expected_wr + 1j * expected_wi

    # Sort by real part then imaginary part
    idx_c = np.lexsort((eig_computed.imag, eig_computed.real))
    idx_e = np.lexsort((eig_expected.imag, eig_expected.real))

    assert_allclose(eig_computed[idx_c], eig_expected[idx_e], rtol=1e-3, atol=1e-3)


def test_mb03xz_eigenvalues_only():
    """
    Test eigenvalue-only mode (JOB='E').

    Random seed: 42 (for reproducibility)
    """
    from slicot import mb03xz

    np.random.seed(42)
    n = 3

    # Create Hermitian G and Q
    a = np.random.randn(n, n) + 1j * np.random.randn(n, n)
    a = np.asfortranarray(a)

    g_full = np.random.randn(n, n) + 1j * np.random.randn(n, n)
    g_full = (g_full + g_full.T.conj()) / 2  # Make Hermitian

    q_full = np.random.randn(n, n) + 1j * np.random.randn(n, n)
    q_full = (q_full + q_full.T.conj()) / 2  # Make Hermitian

    # Pack into QG format: lower tri of Q in cols 0:n-1, upper tri of G in cols 1:n
    qg = np.zeros((n, n+1), dtype=complex, order='F')
    for i in range(n):
        for j in range(i+1):  # Lower triangular Q
            qg[i, j] = q_full[i, j]
        for j in range(i, n):  # Upper triangular G (in cols 1:n)
            qg[i, j+1] = g_full[i, j]

    wr, wi, sc, gc, u1, u2, ilo, scale, info = mb03xz(
        a, qg, balanc='N', job='E', jobu='N'
    )

    assert info == 0
    assert len(wr) == 2 * n
    assert len(wi) == 2 * n

    # Eigenvalues of Hamiltonian: if lambda is eigenvalue, -conj(lambda) also is
    eig = wr + 1j * wi
    for lam in eig:
        neg_conj = -np.conj(lam)
        # Check if -conj(lambda) is also in eigenvalues
        dists = np.abs(eig - neg_conj)
        assert np.min(dists) < 1e-10, f"Missing paired eigenvalue for {lam}"


def test_mb03xz_hamiltonian_structure():
    """
    Validate Hamiltonian eigenvalue pairing: if lambda is an eigenvalue,
    then -conj(lambda) is also an eigenvalue.

    Random seed: 123 (for reproducibility)
    """
    from slicot import mb03xz

    np.random.seed(123)
    n = 4

    # Create random complex Hamiltonian
    a = np.random.randn(n, n) + 1j * np.random.randn(n, n)
    a = np.asfortranarray(a)

    g_full = np.random.randn(n, n) + 1j * np.random.randn(n, n)
    g_full = (g_full + g_full.T.conj()) / 2

    q_full = np.random.randn(n, n) + 1j * np.random.randn(n, n)
    q_full = (q_full + q_full.T.conj()) / 2

    qg = np.zeros((n, n+1), dtype=complex, order='F')
    for i in range(n):
        for j in range(i+1):
            qg[i, j] = q_full[i, j]
        for j in range(i, n):
            qg[i, j+1] = g_full[i, j]

    wr, wi, sc, gc, u1, u2, ilo, scale, info = mb03xz(
        a, qg, balanc='N', job='E', jobu='N'
    )

    assert info == 0

    # Verify Hamiltonian pairing
    eig = wr + 1j * wi
    for lam in eig:
        neg_conj = -np.conj(lam)
        dists = np.abs(eig - neg_conj)
        assert np.min(dists) < 1e-10, f"Eigenvalue {lam} missing pair -conj(lambda)={neg_conj}"


def test_mb03xz_with_balancing():
    """
    Test with balancing enabled (BALANC='B').

    Random seed: 456 (for reproducibility)
    """
    from slicot import mb03xz

    np.random.seed(456)
    n = 3

    # Create poorly conditioned matrix that benefits from balancing
    a = np.array([
        [1e-3+1e-3j, 1e2+0j, 1e1+1e1j],
        [1e-4+0j, 1e3+1e2j, 1e0+0j],
        [1e-2+1e-2j, 1e1+1e1j, 1e-1+1e-1j],
    ], dtype=complex, order='F')

    g_full = np.array([
        [1.0+0j, 0.5+0.1j, 0.2+0.3j],
        [0.5-0.1j, 2.0+0j, 0.4+0.2j],
        [0.2-0.3j, 0.4-0.2j, 0.5+0j],
    ], dtype=complex, order='F')

    q_full = np.array([
        [0.3+0j, 0.1+0.05j, 0.2+0.1j],
        [0.1-0.05j, 0.5+0j, 0.15+0.08j],
        [0.2-0.1j, 0.15-0.08j, 0.4+0j],
    ], dtype=complex, order='F')

    qg = np.zeros((n, n+1), dtype=complex, order='F')
    for i in range(n):
        for j in range(i+1):
            qg[i, j] = q_full[i, j]
        for j in range(i, n):
            qg[i, j+1] = g_full[i, j]

    wr, wi, sc, gc, u1, u2, ilo, scale, info = mb03xz(
        a, qg, balanc='B', job='E', jobu='N'
    )

    assert info == 0
    assert len(wr) == 2 * n
    assert ilo >= 1  # ILO is set by balancing


def test_mb03xz_schur_form():
    """
    Test Schur form computation (JOB='S').

    When JOB='S', the routine returns the complex Schur matrix Sc.
    Random seed: 789 (for reproducibility)
    """
    from slicot import mb03xz

    np.random.seed(789)
    n = 2

    a = np.random.randn(n, n) + 1j * np.random.randn(n, n)
    a = np.asfortranarray(a)

    g_full = np.random.randn(n, n) + 1j * np.random.randn(n, n)
    g_full = (g_full + g_full.T.conj()) / 2

    q_full = np.random.randn(n, n) + 1j * np.random.randn(n, n)
    q_full = (q_full + q_full.T.conj()) / 2

    qg = np.zeros((n, n+1), dtype=complex, order='F')
    for i in range(n):
        for j in range(i+1):
            qg[i, j] = q_full[i, j]
        for j in range(i, n):
            qg[i, j+1] = g_full[i, j]

    wr, wi, sc, gc, u1, u2, ilo, scale, info = mb03xz(
        a, qg, balanc='N', job='S', jobu='N'
    )

    assert info == 0
    assert sc.shape == (2*n, 2*n)

    # Sc should be upper triangular (complex Schur form)
    # Check that lower triangular part is zero
    for i in range(1, 2*n):
        for j in range(i):
            assert np.abs(sc[i, j]) < 1e-10, f"sc[{i},{j}]={sc[i,j]} should be zero"


def test_mb03xz_full_decomposition():
    """
    Test full decomposition (JOB='G', JOBU='U').

    Verifies the transformation satisfies:
    U^H (i*He) U = [Sc  Gc; 0  -Sc^H]

    Random seed: 111 (for reproducibility)
    """
    from slicot import mb03xz

    np.random.seed(111)
    n = 2

    a = np.random.randn(n, n) + 1j * np.random.randn(n, n)
    a = np.asfortranarray(a)

    g_full = np.random.randn(n, n) + 1j * np.random.randn(n, n)
    g_full = (g_full + g_full.T.conj()) / 2

    q_full = np.random.randn(n, n) + 1j * np.random.randn(n, n)
    q_full = (q_full + q_full.T.conj()) / 2

    qg = np.zeros((n, n+1), dtype=complex, order='F')
    for i in range(n):
        for j in range(i+1):
            qg[i, j] = q_full[i, j]
        for j in range(i, n):
            qg[i, j+1] = g_full[i, j]

    wr, wi, sc, gc, u1, u2, ilo, scale, info = mb03xz(
        a, qg, balanc='N', job='G', jobu='U'
    )

    assert info == 0
    assert sc.shape == (2*n, 2*n)
    assert gc.shape == (2*n, 2*n)
    assert u1.shape == (2*n, 2*n)
    assert u2.shape == (2*n, 2*n)

    # Verify U is unitary symplectic: U = [U1 U2; -U2 U1]
    # Check U^H U = I
    u_top = np.hstack([u1, u2])
    u_bot = np.hstack([-u2, u1])
    u = np.vstack([u_top, u_bot])

    ident = u.T.conj() @ u
    assert_allclose(ident, np.eye(4*n), rtol=1e-12, atol=1e-12)


def test_mb03xz_n_zero():
    """Test with N=0 (quick return)."""
    from slicot import mb03xz

    n = 0
    a = np.zeros((0, 0), dtype=complex, order='F')
    qg = np.zeros((0, 1), dtype=complex, order='F')

    wr, wi, sc, gc, u1, u2, ilo, scale, info = mb03xz(
        a, qg, balanc='N', job='E', jobu='N'
    )

    assert info == 0
    assert len(wr) == 0
    assert len(wi) == 0


def test_mb03xz_invalid_balanc():
    """Test invalid BALANC parameter."""
    from slicot import mb03xz

    n = 2
    a = np.zeros((n, n), dtype=complex, order='F')
    qg = np.zeros((n, n+1), dtype=complex, order='F')

    with pytest.raises(ValueError, match="balanc"):
        mb03xz(a, qg, balanc='X', job='E', jobu='N')


def test_mb03xz_invalid_job():
    """Test invalid JOB parameter."""
    from slicot import mb03xz

    n = 2
    a = np.zeros((n, n), dtype=complex, order='F')
    qg = np.zeros((n, n+1), dtype=complex, order='F')

    with pytest.raises(ValueError, match="job"):
        mb03xz(a, qg, balanc='N', job='X', jobu='N')



"""
Tests for MB02XD - Solve A'*A*X = B using Cholesky factorization.

MB02XD computes the solution to a system of linear equations A'*A*X = B,
where A'*A is a symmetric positive definite matrix, using Cholesky
factorization.

Since MB02XD is a low-level routine without Python wrapper, we test the
underlying C implementation via ctypes.
"""

import ctypes
import numpy as np
import os
import pytest


def get_slicot_lib():
    """Load the SLICOT shared library."""
    import glob
    import sys
    base_dir = os.path.join(os.path.dirname(__file__), '..', '..')
    if sys.platform == 'darwin':
        build_dirs = [
            os.path.join(base_dir, "build/macos-arm64-debug/src"),
            os.path.join(base_dir, "build/macos-arm64-release/src"),
        ]
    else:
        build_dirs = [
            os.path.join(base_dir, "build/linux-x64-debug/src"),
            os.path.join(base_dir, "build/linux-x64-release/src"),
            os.path.join(base_dir, "build/linux-x64-debug-sanitizers/src"),
        ]
    for bd in build_dirs:
        if os.path.exists(bd):
            libs = glob.glob(os.path.join(bd, 'libslicot.*'))
            if libs:
                return ctypes.CDLL(libs[0])
    pytest.skip("SLICOT library not found")


def test_mb02xd_basic_full_upper():
    """
    Test MB02XD with full storage, upper triangle.

    Solve A'*A*X = B where A is a 4x3 matrix (M=4, N=3).

    Uses FORM='S' (standard form), STOR='F' (full), UPLO='U' (upper).
    """
    lib = get_slicot_lib()

    m, n, nrhs = 4, 3, 2

    np.random.seed(42)
    A = np.array([
        [1.0, 2.0, 3.0],
        [4.0, 5.0, 6.0],
        [7.0, 8.0, 10.0],
        [1.0, 1.0, 2.0]
    ], order='F', dtype=np.float64)

    B = np.array([
        [1.0, 2.0],
        [3.0, 4.0],
        [5.0, 6.0]
    ], order='F', dtype=np.float64)

    AtA = A.T @ A
    X_expected = np.linalg.solve(AtA, B)

    lda = m
    ldb = n
    ldata = n
    ata = np.zeros((n, n), order='F', dtype=np.float64)
    dwork = np.zeros(1, dtype=np.float64)
    ldwork = 1

    info = ctypes.c_int(0)

    lib.mb02xd.argtypes = [
        ctypes.c_char_p,  # form
        ctypes.c_char_p,  # stor
        ctypes.c_char_p,  # uplo
        ctypes.c_void_p,  # f (callback, NULL for FORM='S')
        ctypes.POINTER(ctypes.c_int),  # m
        ctypes.POINTER(ctypes.c_int),  # n
        ctypes.POINTER(ctypes.c_int),  # nrhs
        ctypes.c_void_p,  # ipar (not used for FORM='S')
        ctypes.POINTER(ctypes.c_int),  # lipar
        ctypes.c_void_p,  # dpar (not used for FORM='S')
        ctypes.POINTER(ctypes.c_int),  # ldpar
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),  # a
        ctypes.POINTER(ctypes.c_int),  # lda
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),  # b
        ctypes.POINTER(ctypes.c_int),  # ldb
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),  # ata
        ctypes.POINTER(ctypes.c_int),  # ldata
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=1, flags='C_CONTIGUOUS'),  # dwork
        ctypes.POINTER(ctypes.c_int),  # ldwork
        ctypes.POINTER(ctypes.c_int),  # info
    ]
    lib.mb02xd.restype = None

    m_c = ctypes.c_int(m)
    n_c = ctypes.c_int(n)
    nrhs_c = ctypes.c_int(nrhs)
    lipar_c = ctypes.c_int(0)
    ldpar_c = ctypes.c_int(0)
    lda_c = ctypes.c_int(lda)
    ldb_c = ctypes.c_int(ldb)
    ldata_c = ctypes.c_int(ldata)
    ldwork_c = ctypes.c_int(ldwork)

    lib.mb02xd(
        b"S",  # form - standard
        b"F",  # stor - full
        b"U",  # uplo - upper
        None,  # f callback (not used)
        ctypes.byref(m_c),
        ctypes.byref(n_c),
        ctypes.byref(nrhs_c),
        None,  # ipar
        ctypes.byref(lipar_c),
        None,  # dpar
        ctypes.byref(ldpar_c),
        A,
        ctypes.byref(lda_c),
        B,
        ctypes.byref(ldb_c),
        ata,
        ctypes.byref(ldata_c),
        dwork,
        ctypes.byref(ldwork_c),
        ctypes.byref(info)
    )

    assert info.value == 0, f"MB02XD failed with info={info.value}"
    np.testing.assert_allclose(B, X_expected, rtol=1e-13, atol=1e-14)

    L_expected = np.linalg.cholesky(AtA).T
    np.testing.assert_allclose(np.triu(ata), L_expected, rtol=1e-13, atol=1e-14)


def test_mb02xd_full_lower():
    """
    Test MB02XD with full storage, lower triangle.

    Random seed: 123 (for reproducibility)
    """
    lib = get_slicot_lib()

    m, n, nrhs = 5, 3, 1

    np.random.seed(123)
    A = np.random.randn(m, n).astype(np.float64, order='F')
    B = np.random.randn(n, nrhs).astype(np.float64, order='F')

    AtA = A.T @ A
    X_expected = np.linalg.solve(AtA, B)

    lda = m
    ldb = n
    ldata = n
    ata = np.zeros((n, n), order='F', dtype=np.float64)
    dwork = np.zeros(1, dtype=np.float64)

    info = ctypes.c_int(0)

    lib.mb02xd.argtypes = [
        ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int),
        ctypes.POINTER(ctypes.c_int), ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int), ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=1, flags='C_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        ctypes.POINTER(ctypes.c_int),
    ]
    lib.mb02xd.restype = None

    lib.mb02xd(
        b"S", b"F", b"L", None,
        ctypes.byref(ctypes.c_int(m)),
        ctypes.byref(ctypes.c_int(n)),
        ctypes.byref(ctypes.c_int(nrhs)),
        None, ctypes.byref(ctypes.c_int(0)),
        None, ctypes.byref(ctypes.c_int(0)),
        A, ctypes.byref(ctypes.c_int(lda)),
        B, ctypes.byref(ctypes.c_int(ldb)),
        ata, ctypes.byref(ctypes.c_int(ldata)),
        dwork, ctypes.byref(ctypes.c_int(1)),
        ctypes.byref(info)
    )

    assert info.value == 0, f"MB02XD failed with info={info.value}"
    np.testing.assert_allclose(B, X_expected, rtol=1e-13, atol=1e-14)


def test_mb02xd_packed_upper():
    """
    Test MB02XD with packed storage, upper triangle.

    For packed storage with UPLO='U', the upper triangle is stored
    column by column: A(1,1), A(1,2), A(2,2), A(1,3), A(2,3), A(3,3), ...

    Random seed: 456 (for reproducibility)
    """
    lib = get_slicot_lib()

    m, n, nrhs = 4, 3, 2

    np.random.seed(456)
    A = np.random.randn(m, n).astype(np.float64, order='F')
    B = np.random.randn(n, nrhs).astype(np.float64, order='F')

    AtA = A.T @ A
    X_expected = np.linalg.solve(AtA, B)

    lda = m
    ldb = n
    packed_size = n * (n + 1) // 2
    ata = np.zeros(packed_size, dtype=np.float64)
    dwork = np.zeros(1, dtype=np.float64)

    info = ctypes.c_int(0)

    lib.mb02xd.argtypes = [
        ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int),
        ctypes.POINTER(ctypes.c_int), ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int), ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=1, flags='C_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=1, flags='C_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        ctypes.POINTER(ctypes.c_int),
    ]
    lib.mb02xd.restype = None

    lib.mb02xd(
        b"S", b"P", b"U", None,
        ctypes.byref(ctypes.c_int(m)),
        ctypes.byref(ctypes.c_int(n)),
        ctypes.byref(ctypes.c_int(nrhs)),
        None, ctypes.byref(ctypes.c_int(0)),
        None, ctypes.byref(ctypes.c_int(0)),
        A, ctypes.byref(ctypes.c_int(lda)),
        B, ctypes.byref(ctypes.c_int(ldb)),
        ata, ctypes.byref(ctypes.c_int(1)),
        dwork, ctypes.byref(ctypes.c_int(1)),
        ctypes.byref(info)
    )

    assert info.value == 0, f"MB02XD failed with info={info.value}"
    np.testing.assert_allclose(B, X_expected, rtol=1e-13, atol=1e-14)


def test_mb02xd_packed_lower():
    """
    Test MB02XD with packed storage, lower triangle.

    For packed storage with UPLO='L', the lower triangle is stored
    column by column: A(1,1), A(2,1), A(3,1), ..., A(2,2), A(3,2), ...

    Random seed: 789 (for reproducibility)
    """
    lib = get_slicot_lib()

    m, n, nrhs = 6, 4, 3

    np.random.seed(789)
    A = np.random.randn(m, n).astype(np.float64, order='F')
    B = np.random.randn(n, nrhs).astype(np.float64, order='F')

    AtA = A.T @ A
    X_expected = np.linalg.solve(AtA, B)

    lda = m
    ldb = n
    packed_size = n * (n + 1) // 2
    ata = np.zeros(packed_size, dtype=np.float64)
    dwork = np.zeros(1, dtype=np.float64)

    info = ctypes.c_int(0)

    lib.mb02xd.argtypes = [
        ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int),
        ctypes.POINTER(ctypes.c_int), ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int), ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=1, flags='C_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=1, flags='C_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        ctypes.POINTER(ctypes.c_int),
    ]
    lib.mb02xd.restype = None

    lib.mb02xd(
        b"S", b"P", b"L", None,
        ctypes.byref(ctypes.c_int(m)),
        ctypes.byref(ctypes.c_int(n)),
        ctypes.byref(ctypes.c_int(nrhs)),
        None, ctypes.byref(ctypes.c_int(0)),
        None, ctypes.byref(ctypes.c_int(0)),
        A, ctypes.byref(ctypes.c_int(lda)),
        B, ctypes.byref(ctypes.c_int(ldb)),
        ata, ctypes.byref(ctypes.c_int(1)),
        dwork, ctypes.byref(ctypes.c_int(1)),
        ctypes.byref(info)
    )

    assert info.value == 0, f"MB02XD failed with info={info.value}"
    np.testing.assert_allclose(B, X_expected, rtol=1e-13, atol=1e-14)


def test_mb02xd_singular_matrix():
    """
    Test MB02XD with singular A'*A matrix (rank deficient A).

    Should return info > 0 indicating singularity.
    """
    lib = get_slicot_lib()

    m, n, nrhs = 4, 3, 1

    A = np.array([
        [1.0, 2.0, 3.0],
        [2.0, 4.0, 6.0],
        [1.0, 2.0, 3.0],
        [0.0, 0.0, 0.0]
    ], order='F', dtype=np.float64)

    B = np.array([[1.0], [2.0], [3.0]], order='F', dtype=np.float64)

    lda = m
    ldb = n
    ldata = n
    ata = np.zeros((n, n), order='F', dtype=np.float64)
    dwork = np.zeros(1, dtype=np.float64)

    info = ctypes.c_int(0)

    lib.mb02xd.argtypes = [
        ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int),
        ctypes.POINTER(ctypes.c_int), ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int), ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=1, flags='C_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        ctypes.POINTER(ctypes.c_int),
    ]
    lib.mb02xd.restype = None

    lib.mb02xd(
        b"S", b"F", b"U", None,
        ctypes.byref(ctypes.c_int(m)),
        ctypes.byref(ctypes.c_int(n)),
        ctypes.byref(ctypes.c_int(nrhs)),
        None, ctypes.byref(ctypes.c_int(0)),
        None, ctypes.byref(ctypes.c_int(0)),
        A, ctypes.byref(ctypes.c_int(lda)),
        B, ctypes.byref(ctypes.c_int(ldb)),
        ata, ctypes.byref(ctypes.c_int(ldata)),
        dwork, ctypes.byref(ctypes.c_int(1)),
        ctypes.byref(info)
    )

    assert info.value > 0, f"Expected info > 0 for singular matrix, got {info.value}"


def test_mb02xd_parameter_errors():
    """
    Test MB02XD parameter validation.
    """
    lib = get_slicot_lib()

    m, n, nrhs = 4, 3, 2
    A = np.ones((m, n), order='F', dtype=np.float64)
    B = np.ones((n, nrhs), order='F', dtype=np.float64)
    ata = np.zeros((n, n), order='F', dtype=np.float64)
    dwork = np.zeros(1, dtype=np.float64)

    lib.mb02xd.argtypes = [
        ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int),
        ctypes.POINTER(ctypes.c_int), ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int), ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=1, flags='C_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        ctypes.POINTER(ctypes.c_int),
    ]
    lib.mb02xd.restype = None

    info = ctypes.c_int(0)
    lib.mb02xd(
        b"X", b"F", b"U", None,
        ctypes.byref(ctypes.c_int(m)),
        ctypes.byref(ctypes.c_int(n)),
        ctypes.byref(ctypes.c_int(nrhs)),
        None, ctypes.byref(ctypes.c_int(0)),
        None, ctypes.byref(ctypes.c_int(0)),
        A, ctypes.byref(ctypes.c_int(m)),
        B, ctypes.byref(ctypes.c_int(n)),
        ata, ctypes.byref(ctypes.c_int(n)),
        dwork, ctypes.byref(ctypes.c_int(1)),
        ctypes.byref(info)
    )
    assert info.value == -1, f"Expected info=-1 for invalid FORM, got {info.value}"

    info = ctypes.c_int(0)
    lib.mb02xd(
        b"S", b"X", b"U", None,
        ctypes.byref(ctypes.c_int(m)),
        ctypes.byref(ctypes.c_int(n)),
        ctypes.byref(ctypes.c_int(nrhs)),
        None, ctypes.byref(ctypes.c_int(0)),
        None, ctypes.byref(ctypes.c_int(0)),
        A, ctypes.byref(ctypes.c_int(m)),
        B, ctypes.byref(ctypes.c_int(n)),
        ata, ctypes.byref(ctypes.c_int(n)),
        dwork, ctypes.byref(ctypes.c_int(1)),
        ctypes.byref(info)
    )
    assert info.value == -2, f"Expected info=-2 for invalid STOR, got {info.value}"

    info = ctypes.c_int(0)
    lib.mb02xd(
        b"S", b"F", b"U", None,
        ctypes.byref(ctypes.c_int(-1)),
        ctypes.byref(ctypes.c_int(n)),
        ctypes.byref(ctypes.c_int(nrhs)),
        None, ctypes.byref(ctypes.c_int(0)),
        None, ctypes.byref(ctypes.c_int(0)),
        A, ctypes.byref(ctypes.c_int(m)),
        B, ctypes.byref(ctypes.c_int(n)),
        ata, ctypes.byref(ctypes.c_int(n)),
        dwork, ctypes.byref(ctypes.c_int(1)),
        ctypes.byref(info)
    )
    assert info.value == -5, f"Expected info=-5 for M<0, got {info.value}"


def test_mb02xd_residual_property():
    """
    Mathematical property test: verify residual A'*A*X - B is small.

    This validates that the solution X truly satisfies the normal equations.

    Random seed: 999 (for reproducibility)
    """
    lib = get_slicot_lib()

    m, n, nrhs = 10, 5, 3

    np.random.seed(999)
    A = np.random.randn(m, n).astype(np.float64, order='F')
    B_orig = np.random.randn(n, nrhs).astype(np.float64, order='F')
    B = B_orig.copy(order='F')

    AtA = A.T @ A

    lda = m
    ldb = n
    ldata = n
    ata = np.zeros((n, n), order='F', dtype=np.float64)
    dwork = np.zeros(1, dtype=np.float64)

    info = ctypes.c_int(0)

    lib.mb02xd.argtypes = [
        ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int),
        ctypes.POINTER(ctypes.c_int), ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int), ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='F_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        np.ctypeslib.ndpointer(dtype=np.float64, ndim=1, flags='C_CONTIGUOUS'),
        ctypes.POINTER(ctypes.c_int),
        ctypes.POINTER(ctypes.c_int),
    ]
    lib.mb02xd.restype = None

    lib.mb02xd(
        b"S", b"F", b"U", None,
        ctypes.byref(ctypes.c_int(m)),
        ctypes.byref(ctypes.c_int(n)),
        ctypes.byref(ctypes.c_int(nrhs)),
        None, ctypes.byref(ctypes.c_int(0)),
        None, ctypes.byref(ctypes.c_int(0)),
        A, ctypes.byref(ctypes.c_int(lda)),
        B, ctypes.byref(ctypes.c_int(ldb)),
        ata, ctypes.byref(ctypes.c_int(ldata)),
        dwork, ctypes.byref(ctypes.c_int(1)),
        ctypes.byref(info)
    )

    assert info.value == 0, f"MB02XD failed with info={info.value}"

    X = B
    residual = AtA @ X - B_orig

    residual_norm = np.linalg.norm(residual, 'fro')
    B_norm = np.linalg.norm(B_orig, 'fro')

    relative_residual = residual_norm / B_norm
    assert relative_residual < 1e-13, f"Relative residual {relative_residual} too large"


if __name__ == "__main__":
    pytest.main([__file__, "-v"])
